"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.

    Module with classes and methods to perform implicit regional modelling based on
    the potential field method.
    Tested on Ubuntu 16

    Created on 10/04/2018

    @author: Elisa Heim, Miguel de la Varga
"""

# This is for sphenix to find the packages
# sys.path.append( path.dirname( path.dirname( path.abspath(__file__) ) ) )

from typing import Set, Tuple, Dict, Union
import gempy_legacy as _gempy
from ._visualization_2d import PlotData2D, PlotSolution
from .visualization_3d import GemPyvtkInteract


def plot_data_3D(geo_data, ve=1, **kwargs):
    """
    Plot in vtk all the input data of a model
    Args:
        geo_data (gempy.DataManagement.InputData): Input data of the model

    Returns:
        None
    """
    vv = GemPyvtkInteract(geo_data, ve=ve, **kwargs)
    # vv.restart()
    vv.set_surface_points()
    vv.set_orientations()
    vv.render_model(**kwargs)

    return vv


def plot_3D(geo_model, render_surfaces=True, render_data=True,
            render_topography=True,
            real_time=False, **kwargs):
    """
        Plot in vtk all the input data of a model
        Args:
            geo_model (gempy.DataManagement.InputData): Input data of the model

        Returns:
            None
        """
    vv = GemPyvtkInteract(geo_model, real_time=real_time, **kwargs)
    # vv.restart()
    if render_data is True:
        vv.set_surface_points()
        vv.set_orientations()
    if render_surfaces is True:
        vv.set_surfaces(geo_model._surfaces)
    if render_topography is True and geo_model._grid.topography is not None:
        vv.set_topography()

    vv.render_model(**kwargs)

    return vv


def export_to_vtk(geo_data, path=None, name=None, voxels=True, block=None, surfaces=True):
    """
      Export data to a vtk file for posterior visualizations

      Args:
          geo_data(:class:`Model`)
          path(str): path to the location of the vtk
          name(str): Name of the files. Default name: Default
          voxels(bool): if True export lith_block
          block(Optional[np.array]): One of the solutions of the regular grid. This can be used if for
           example you want to export an scalar field or an specific series block. If None is passed, lith_block
           will be exported.
          surfaces(bool): If True, export the polydata surfaces.

      Returns:
          None
      """

    if voxels is True:
        GemPyvtkInteract.export_vtk_lith_block(geo_data, lith_block=block,
                                               path=path)
    if surfaces is True:
        geo_data.solutions.compute_all_surfaces()
        ver, sim = _gempy.get_surfaces(geo_data)
        GemPyvtkInteract.export_vtk_surfaces(geo_data, ver, sim, path=path,
                                             name=name)
    return True


def plot_data(geo_data, direction="y", data_type='all', series="all",
              show_legend=True, **kwargs):
    """
    Plot the projection of the raw data (surface_points and orientations) in 2D following a
    specific directions

    Args:
        direction(str): xyz. Caartesian direction to be plotted
        series(str): series to plot
        ve(float): Vertical exageration
        **kwargs: seaborn lmplot key arguments. (TODO: adding the link to them)

    Returns:
        None
    """
    plot = PlotData2D(geo_data)
    p = plot.plot_data(direction=direction, data_type=data_type, series=series,
                       show_legend=show_legend, **kwargs)
    # TODO saving options
    return plot


def plot_stereonet(geo_data, litho=None, planes=True, poles=True,
                   single_plots=False,
                   show_density=False):
    '''
    Plot an equal-area projection of the orientations dataframe using mplstereonet.

    Args:
        geo_model (gempy.DataManagement.InputData): Input data of the model
        series_only: To select whether a stereonet is plotted perries or per formation
        litho: selection of formation or series names, as list. If None, all are plotted
        planes: If True, azimuth and dip are plotted as great circles
        poles: If True, pole points (plane normal vectors) of azimuth and dip are plotted
        single_plots: If True, each formation is plotted in a single stereonet
        show_density: If True, density contour plot around the pole points is shown

    Returns:
        None
    '''

    plot = PlotData2D(geo_data)
    plot.plot_stereonet(litho=litho, planes=planes, poles=poles,
                        single_plots=single_plots,
                        show_density=show_density)


def plot_map(model, contour_lines=True, show_data=True, show_hillshades: bool = False, figsize=(12, 12), **kwargs):
    """

    Args:
        figsize:
        show_hillshades:
        model:
        contour_lines:
        show_faults:
        show_data:
        **kwargs

    Returns:

    """
    plot = PlotSolution(model)
    plot.plot_map(contour_lines=contour_lines, show_data=show_data, show_hillshades=show_hillshades,
                  figsize=figsize, **kwargs)


def plot_section_traces(model, section_names=None, contour_lines=False,
                        show_data=True, show_all_data=False):
    """

    Args:
        model:
        show_data:
        section_names:
        contour_lines:

    Returns:

    """
    plot = PlotSolution(model)
    if plot.model.solutions.geological_map is not None:
        plot.plot_map(contour_lines=contour_lines, show_data=show_data,
                      show_all_data=show_all_data)
    # else:
    # fig = plt.figure()
    # plt.title('Section traces, z direction')

    plot.plot_section_traces(show_data=show_data, section_names=section_names,
                             contour_lines=contour_lines,
                             show_all_data=show_all_data)


"""
def plot_predef_sections(model, show_traces=True, show_data=False, section_names=None, show_faults=True,
                         show_topo=True, figsize=(12, 12)):


    Args:
        model:
        show_traces:
        show_data:
        section_names:
        show_faults:
        show_topo:
        figsize:

    Returns:


    plot = PlotSolution(model)
    plot.plot_sections(show_traces=show_traces, show_data=show_data, section_names=section_names,
                       show_faults=show_faults, show_topo=show_topo, figsize=figsize)

"""


def plot_section_by_name(model, section_name, show_faults=True, show_topo=True,
                         show_data=True,
                         show_all_data=False, radius='default',
                         contourplot=True):
    # Todo needs more keywords:
    ### if show_data: radius, data_type
    plot = PlotSolution(model)
    plot.plot_section_by_name(section_name=section_name, show_topo=show_topo,
                              show_faults=show_faults,
                              show_data=show_data, show_all_data=show_all_data,
                              radius=radius, contourplot=contourplot)


def plot_all_sections(model, show_data=False, section_names=None,
                      show_topo=True, figsize=(12, 12)):
    plot = PlotSolution(model)
    plot.plot_all_sections(show_data=show_data, section_names=section_names,
                           show_topo=show_topo,
                           figsize=figsize)


def plot_section(model, cell_number=13, block=None, direction="y",
                 interpolation='none',
                 show_data=True, show_faults=True, show_topo=False,
                 block_type=None, ve=1,
                 show_all_data=False, show_legend=True, **kwargs):
    """
    Plot a section of the block model

    Args:
        cell_number(int): position of the array to plot
        direction(str): xyz. Caartesian direction to be plotted
        interpolation(str): Type of interpolation of plt.imshow. Default 'none'.  Acceptable values are 'none'
        ,'nearest', 'bilinear', 'bicubic',
        'spline16', 'spline36', 'hanning', 'hamming', 'hermite', 'kaiser',
        'quadric', 'catrom', 'gaussian', 'bessel', 'mitchell', 'sinc',
        'lanczos'
        ve(float): Vertical exageration
        **kwargs: imshow keywargs

    Returns:
        None
    """
    plot = PlotSolution(model)
    plot.fig = plot.plot_block_section(model.solutions, cell_number, block,
                                       direction, interpolation,
                                       show_data, show_faults, show_topo,
                                       block_type, ve,
                                       show_all_data=show_all_data,
                                       show_legend=show_legend, **kwargs)
    return plot


def plot_scalar_field(model, cell_number, N=20,
                      direction="y", block=None, alpha=0.6, show_data=True,
                      show_all_data=False, series=0, *args, **kwargs):
    """
    Plot a potential field in a given direction.

    Args:
        cell_number(int): position of the array to plot
        potential_field(str): name of the potential field (or series) to plot
        n_pf(int): number of the  potential field (or series) to plot
        direction(str): xyz. Caartesian direction to be plotted
        serie: *Deprecated*
        **kwargs: plt.contour kwargs

    Returns:
        None
    """
    plot = PlotSolution(model)
    if block is not None:
        block = block
    else:
        block = model.solutions

    plot.plot_scalar_field(block, cell_number, N=N,
                           direction=direction, show_data=show_data,
                           series=series, alpha=alpha,
                           show_all_data=show_all_data,
                           *args, **kwargs)


def plot_section_scalarfield(model, section_name, sn, levels=50,
                             show_faults=True, show_topo=True, lithback=True):
    """
    Plot the potential field in the predefined sections.
    Args:
        model:
        section_name: name of the section
        sn: scalar field number, order like in model.series
        levels: number of isolines you want to plot
        show_faults: whether or not faults should be plotted
        show_topo: whether or not the topography should be plotted
        lithback: lithology background

    Returns:
        None
    """
    plot = PlotSolution(model)
    plot.plot_section_scalarfield(section_name=section_name, sn=sn,
                                  levels=levels, show_faults=show_faults,
                                  show_topo=show_topo, lithback=lithback)


def plot_gradient(geo_data, scalar_field, gx, gy, gz, cell_number, q_stepsize=5,
                  direction="y", plot_scalar=True, **kwargs):
    """
        Plot the gradient of the scalar field in a given direction.

        Args:
            geo_data (gempy.DataManagement.InputData): Input data of the model
            scalar_field(numpy.array): scalar field to plot with the gradient
            gx(numpy.array): gradient in x-direction
            gy(numpy.array): gradient in y-direction
            gz(numpy.array): gradient in z-direction
            cell_number(int): position of the array to plot
            q_stepsize(int): step size between arrows to indicate gradient
            direction(str): xyz. Caartesian direction to be plotted
            plot_scalar(bool): boolean to plot scalar field
            **kwargs: plt.contour kwargs

        Returns:
            None
    """
    plot = PlotSolution(geo_data)
    plot.plot_gradient(scalar_field, gx, gy, gz, cell_number,
                       q_stepsize=q_stepsize,
                       direction=direction, plot_scalar=plot_scalar,
                       **kwargs)


def plot_topology(
        geo_model,
        edges: Set[Tuple[int, int]],
        centroids: Dict,
        direction: Union["x", "y", "z"] = "y",
        scale: bool = True,
        label_kwargs: dict = None,
        edge_kwargs: dict = None
):
    """Plot the topology adjacency graph in 2-D.

    Args:
        geo_model ([type]): GemPy geomodel instance.
        edges (Set[Tuple[int, int]]): Set of topology edges.
        centroids (Dict[int, Array[int, 3]]): Dictionary of topology id's and
            their centroids.
        direction (Union["x", "y", "z", optional): Section direction.
            Defaults to "y".
        label_kwargs (dict, optional): Keyword arguments for topology labels.
            Defaults to None.
        edge_kwargs (dict, optional): Keyword arguments for topology edges.
            Defaults to None.
    """
    PlotSolution.plot_topo_g(
        geo_model,
        edges,
        centroids,
        direction=direction,
        scale=scale,
        label_kwargs=label_kwargs,
        edge_kwargs=edge_kwargs
    )


def plot_ar(geo_model, path=None, project_name=None, api_token=None, secret=None):
    """ Create, upload and retrieve tag to visualize the model in AR in rexview

    https://www.rexos.org/getting-started/

    Args:
        geo_model (gempy.Model):
        path: Location for rex files. Default cwd
        project_name: Name of the project in rexos
        api_token: rexos api token
        secret: rexos secret

    Returns:
        gempy.addons.rex_api.Rextag
    """
    from gempy_legacy.addons.rex_api import upload_to_rexcloud
    from gempy_legacy.addons.gempy_to_rexfile import write_rex, geomodel_to_rex
    if project_name is None:
        project_name = geo_model.meta.project_name

    if path is None:
        path = './'

    rex_bytes = geomodel_to_rex(geo_model)
    files_path = write_rex(rex_bytes, path)
    project_name_ = project_name
    for i in range(40):
        try:
            tag = upload_to_rexcloud(files_path, project_name=project_name_, api_token=api_token, secret=secret)
            break
        except ConnectionError:
            project_name_ = project_name + str(i)
            pass

    return tag



from typing import Union
from .data import AdditionalData
from .data_modules.stack import Faults, Series
from ..utils.meta import _setdoc_pro, _setdoc
import gempy_legacy.utils.docstring as ds
from .surfaces import Surfaces
from .grid import Grid
from .data_modules.surface_points import SurfacePoints
from .data_modules.orientations import Orientations

import numpy as np
import aesara


@_setdoc_pro(
    [SurfacePoints.__doc__, Orientations.__doc__, Grid.__doc__, Surfaces.__doc__, Series.__doc__,
     Faults.__doc__, AdditionalData.__doc__])
class Interpolator(object):
    """Class that act as:
     1) linker between the data objects and the aesara graph
     2) container of aesara graphs + shared variables
     3) container of aesara function

    Args:
        surface_points (SurfacePoints): [s0]
        orientations (Orientations): [s1]
        grid (Grid): [s2]
        surfaces (Surfaces): [s3]
        series (Series): [s4]
        faults (Faults): [s5]
        additional_data (AdditionalData): [s6]
        kwargs:
            - compile_aesara: if true, the function is compile at the creation of the class

    Attributes:
        surface_points (SurfacePoints)
        orientations (Orientations)
        grid (Grid)
        surfaces (Surfaces)
        faults (Faults)
        additional_data (AdditionalData)
        dtype (['float32', 'float64']): float precision
        aesara_graph: aesara graph object with the properties from AdditionalData -> Options
        aesara function: python function to call the aesara code

    """

    # TODO assert passed data is rescaled

    def __init__(self, surface_points: "SurfacePoints", orientations: "Orientations", grid: "Grid",
                 surfaces: "Surfaces", series: Series, faults: "Faults",
                 additional_data: "AdditionalData", **kwargs):
        # Test
        self.surface_points = surface_points
        self.orientations = orientations
        self.grid = grid
        self.additional_data = additional_data
        self.surfaces = surfaces
        self.series = series
        self.faults = faults

        self.dtype = additional_data.options.df.loc['values', 'dtype']
        self.aesara_graph = self.create_aesara_graph(additional_data, inplace=False)
        self.aesara_function = None

        self._compute_len_series()

    def _compute_len_series(self):
        self.len_series_i = self.additional_data.structure_data.df.loc[
                                'values', 'len series surface_points'] - \
                            self.additional_data.structure_data.df.loc[
                                'values', 'number surfaces per series']
        if self.len_series_i.shape[0] == 0:
            self.len_series_i = np.zeros(1, dtype=int)

        self.len_series_o = self.additional_data.structure_data.df.loc[
            'values', 'len series orientations'].astype(
            'int32')
        if self.len_series_o.shape[0] == 0:
            self.len_series_o = np.zeros(1, dtype=int)

        self.len_series_u = self.additional_data.kriging_data.df.loc[
            'values', 'drift equations'].astype('int32')
        if self.len_series_u.shape[0] == 0:
            self.len_series_u = np.zeros(1, dtype=int)

        self.len_series_f = self.faults.faults_relations_df.sum(axis=0).values.astype('int32')[:self.additional_data.get_additional_data()['values']['Structure', 'number series']]
        if self.len_series_f.shape[0] == 0:
            self.len_series_f = np.zeros(1, dtype=int)

        self.len_series_w = self.len_series_i + self.len_series_o * 3 + self.len_series_u + self.len_series_f

    @_setdoc_pro([AdditionalData.__doc__, ds.inplace, ds.aesara_graph_pro])
    def create_aesara_graph(self, additional_data: "AdditionalData" = None, inplace=True,
                            output=None, **kwargs):
        """
        Create the graph accordingly to the options in the AdditionalData object

        Args:
            additional_data (AdditionalData): [s0]
            inplace (bool): [s1]

        Returns:
            aesaraGraphPro: [s2]
        """
        if output is None:
            output = ['geology']

        import gempy_legacy.core.aesara_modules.aesara_graph_pro as tg
        import importlib
        importlib.reload(tg)

        if additional_data is None:
            additional_data = self.additional_data

        self.dtype = additional_data.options.df.loc['values', 'dtype']

        graph = tg.aesaraGraphPro(
            optimizer=additional_data.options.df.loc['values', 'aesara_optimizer'],
            verbose=additional_data.options.df.loc['values', 'verbosity'],
            output=output,
            **kwargs)
        if inplace is True:
            self.aesara_graph = graph
        else:
            return graph

    @_setdoc_pro([ds.aesara_graph_pro])
    def set_aesara_graph(self, th_graph):
        """
        Attach an already create aesara graph.

        Args:
            th_graph (aesaraGraphPro): [s0]

        Returns:
            True
        """
        self.aesara_graph = th_graph
        return True

    def set_aesara_shared_kriging(self):
        """
        Set to the aesara_graph attribute the shared variables of kriging values from the linked
         :class:`AdditionalData`.

        Returns:
            True
        """
        # Range
        range_val = self.additional_data.kriging_data.df.loc['values', 'range']
        range_res = range_val / self.additional_data.rescaling_data.df.loc[
            'values', 'rescaling factor']

        range_list = range_res * np.ones(
            self.additional_data.structure_data.df.loc['values',
            'number series'])
        #
        # if type(range_res) is np.float or type(range_res) is np.float64:
        #     range_list = range_res * np.ones(
        #         self.additional_data.structure_data.df.loc['values',
        #                                                    'number series']
        #     )
        # elif isinstance(range_res, Iterable):
        #     range_list = range_res
        # else:
        #     raise AttributeError('Range must be either int or Iterable')

        # TODO add rescaled range and co into the rescaling data df?
        self.aesara_graph.a_T.set_value(np.cast[self.dtype](range_list))

        # Covariance at 0

        cov_val = self.additional_data.kriging_data.df.loc['values', '$C_o$']
        cov_res = cov_val / self.additional_data.rescaling_data.df.loc[
            'values', 'rescaling factor']
        cov_list = cov_res * np.ones(
            self.additional_data.structure_data.df.loc['values',
            'number series']
        )
        #
        #
        # if type(cov_res) is np.float or type(cov_res) is np.float64:
        #     cov_list = cov_res * np.ones(
        #         self.additional_data.structure_data.df.loc['values',
        #                                                    'number series']
        #     )
        # elif isinstance(cov_res, Iterable):
        #     cov_list = cov_res
        # else:
        #     raise AttributeError('Covariance at 0 must be either int or Iterable')

        self.aesara_graph.c_o_T.set_value(
            np.cast[self.dtype](cov_list)
        )

        # universal grades
        self.aesara_graph.n_universal_eq_T.set_value(
            list(self.additional_data.kriging_data.df.loc['values', 'drift equations'].astype(
                'int32')[self.non_zero]))

        self.set_aesara_shared_nuggets()

    def set_aesara_shared_nuggets(self):
        # nugget effect
        # len_orientations = self.additional_data.structure_data.df.loc['values', 'len series orientations']
        # len_orientations_len = np.sum(len_orientations)

        self.aesara_graph.nugget_effect_grad_T.set_value(
            np.cast[self.dtype](np.tile(
                self.orientations.df['smooth'], 3)))

        # len_rest_form = (self.additional_data.structure_data.df.loc['values', 'len surfaces surface_points'])
        # len_rest_len = np.sum(len_rest_form)
        self.aesara_graph.nugget_effect_scalar_T.set_value(
            np.cast[self.dtype](self.surface_points.df['smooth']))
        return True

    def set_aesara_shared_structure_surfaces(self):
        """
        Set to the aesara_graph attribute the shared variables of structure from the linked
         :class:`AdditionalData`.

        Returns:
            True
        """
        len_rest_form = (self.additional_data.structure_data.df.loc[
                             'values', 'len surfaces surface_points'] - 1)
        self.aesara_graph.number_of_points_per_surface_T.set_value(len_rest_form.astype('int32'))


class InterpolatorWeights(Interpolator):

    def __init__(self, surface_points: "SurfacePoints", orientations: "Orientations", grid: "Grid",
                 surfaces: "Surfaces", series, faults: "Faults", additional_data: "AdditionalData",
                 **kwargs):

        super(InterpolatorWeights, self).__init__(surface_points, orientations, grid, surfaces,
                                                  series, faults,
                                                  additional_data, **kwargs)

    def get_python_input_weights(self, fault_drift=None):
        """
             Get values from the data objects used during the interpolation:
                 - dip positions XYZ
                 - dip angles
                 - azimuth
                 - polarity
                 - surface_points coordinates XYZ
             Returns:
                 (list)
             """
        # orientations, this ones I tile them inside aesara. PYTHON VAR
        dips_position = self.orientations.df[['X_c', 'Y_c', 'Z_c']].values
        dip_angles = self.orientations.df["dip"].values
        azimuth = self.orientations.df["azimuth"].values
        polarity = self.orientations.df["polarity"].values
        surface_points_coord = self.surface_points.df[['X_c', 'Y_c', 'Z_c']].values
        if fault_drift is None:
            fault_drift = np.zeros((0, self.grid.values.shape[0] + 2 * self.len_series_i.sum()))

        #  fault_drift = np.zeros((0, surface_points_coord.shape[0]))

        # Set all in a list casting them in the chosen dtype
        idl = [np.cast[self.dtype](xs) for xs in
               (dips_position, dip_angles, azimuth, polarity, surface_points_coord,
                fault_drift)]
        return idl

    def compile_th_fn(self, inplace=False, debug=False):

        self.set_aesara_shared_kriging()
        self.set_aesara_shared_structure_surfaces()
        # This are the shared parameters and the compilation of the function. This will be hidden as well at some point
        input_data_T = self.aesara_graph.input_parameters_kriging
        print('Compiling aesara function...')

        th_fn = aesara.function(input_data_T,
                                self.aesara_graph.compute_weights(),
                                # mode=NanGuardMode(nan_is_error=True),
                                on_unused_input='warn',
                                allow_input_downcast=False,
                                profile=False)
        if inplace is True:
            self.aesara_function = th_fn

        if debug is True:
            print('Level of Optimization: ', aesara.config.optimizer)
            print('Device: ', aesara.config.device)
            print('Precision: ', self.dtype)
            print('Number of faults: ',
                  self.additional_data.structure_data.df.loc['values', 'number faults'])
        print('Compilation Done!')
        return th_fn


class InterpolatorScalar(Interpolator):

    def __init__(self, surface_points: "SurfacePoints", orientations: "Orientations", grid: "Grid",
                 surfaces: "Surfaces", series, faults: "Faults", additional_data: "AdditionalData",
                 **kwargs):

        super(InterpolatorScalar, self).__init__(surface_points, orientations, grid, surfaces,
                                                 series, faults,
                                                 additional_data, **kwargs)

    def get_python_input_zx(self, fault_drift=None):
        """
             Get values from the data objects used during the interpolation:
                 - dip positions XYZ
                 - dip angles
                 - azimuth
                 - polarity
                 - surface_points coordinates XYZ
             Returns:
                 (list)
             """
        # orientations, this ones I tile them inside aesara. PYTHON VAR
        dips_position = self.orientations.df[['X_c', 'Y_c', 'Z_c']].values
        dip_angles = self.orientations.df["dip"].values
        azimuth = self.orientations.df["azimuth"].values
        polarity = self.orientations.df["polarity"].values
        surface_points_coord = self.surface_points.df[['X_c', 'Y_c', 'Z_c']].values
        grid = self.grid.values_c

        if fault_drift is None:
            fault_drift = np.zeros((0, grid.shape[0] + 2 * self.len_series_i.sum()))

        #      fault_drift = np.zeros((0, grid.shape[0] + surface_points_coord.shape[0]))

        # Set all in a list casting them in the chosen dtype
        idl = [np.cast[self.dtype](xs) for xs in
               (dips_position, dip_angles, azimuth, polarity, surface_points_coord,
                fault_drift, grid)]
        return idl

    def compile_th_fn(self, weights=None, grid=None, inplace=False, debug=False):
        """

        Args:
            weights: Constant weights
            grid:  Constant grids
            inplace:
            debug:

        Returns:

        """
        self.set_aesara_shared_kriging()
        self.set_aesara_shared_structure_surfaces()
        # This are the shared parameters and the compilation of the function. This will be hidden as well at some point
        input_data_T = self.aesara_graph.input_parameters_kriging_export
        print('Compiling aesara function...')

        if weights is None:
            weights = self.aesara_graph.compute_weights()
        else:
            weights = aesara.shared(weights)

        if grid is None:
            grid = self.aesara_graph.grid_val_T
        else:
            grid = aesara.shared(grid)

        th_fn = aesara.function(input_data_T,
                                self.aesara_graph.compute_scalar_field(weights, grid),
                                # mode=NanGuardMode(nan_is_error=True),
                                on_unused_input='ignore',
                                allow_input_downcast=False,
                                profile=False)

        if inplace is True:
            self.aesara_function = th_fn

        if debug is True:
            print('Level of Optimization: ', aesara.config.optimizer)
            print('Device: ', aesara.config.device)
            print('Precision: ', aesara.config.floatX)
            print('Number of faults: ',
                  self.additional_data.structure_data.df.loc['values', 'number faults'])
        print('Compilation Done!')
        return th_fn


class InterpolatorBlock(Interpolator):

    def __init__(self, surface_points: "SurfacePoints", orientations: "Orientations", grid: "Grid",
                 surfaces: "Surfaces", series: Series, faults: "Faults",
                 additional_data: "AdditionalData", **kwargs):

        super(InterpolatorBlock, self).__init__(surface_points, orientations, grid, surfaces,
                                                series,
                                                faults, additional_data, **kwargs)
        self.aesara_function_formation = None
        self.aesara_function_faults = None

    def get_python_input_block(self, fault_drift=None):
        """
             Get values from the data objects used during the interpolation:
                 - dip positions XYZ
                 - dip angles
                 - azimuth
                 - polarity
                 - surface_points coordinates XYZ
             Returns:
                 (list)
             """
        # orientations, this ones I tile them inside aesara. PYTHON VAR
        dips_position = self.orientations.df[['X_c', 'Y_c', 'Z_c']].values
        dip_angles = self.orientations.df["dip"].values
        azimuth = self.orientations.df["azimuth"].values
        polarity = self.orientations.df["polarity"].values
        surface_points_coord = self.surface_points.df[['X_c', 'Y_c', 'Z_c']].values
        grid = self.grid.values_c
        if fault_drift is None:
            fault_drift = np.zeros((0, grid.shape[0] + 2 * self.len_series_i.sum()))

        values_properties = self.surfaces.df.iloc[:, self.surfaces._n_properties:].values.astype(
            self.dtype).T

        # Set all in a list casting them in the chosen dtype
        idl = [np.cast[self.dtype](xs) for xs in
               (dips_position, dip_angles, azimuth, polarity, surface_points_coord,
                fault_drift, grid, values_properties)]
        return idl

    def compile_th_fn_formation_block(self, Z_x=None, weights=None, grid=None,
                                      values_properties=None, inplace=False,
                                      debug=False):
        """

        Args:
            weights: Constant weights
            grid:  Constant grids
            inplace:
            debug:

        Returns:

        """
        self.set_aesara_shared_kriging()
        self.set_aesara_shared_structure_surfaces()
        # This are the shared parameters and the compilation of the function. This will be hidden as well at some point
        input_data_T = self.aesara_graph.input_parameters_block
        print('Compiling aesara function...')

        if weights is None:
            weights = self.aesara_graph.compute_weights()
        else:
            weights = aesara.shared(weights)

        if grid is None:
            grid = self.aesara_graph.grid_val_T
        else:
            grid = aesara.shared(grid)

        if values_properties is None:
            values_properties = self.aesara_graph.values_properties_op
        else:
            values_properties = aesara.shared(values_properties)

        if Z_x is None:
            Z_x = self.aesara_graph.compute_scalar_field(weights, grid)
        else:
            Z_x = aesara.shared(Z_x)

        th_fn = aesara.function(input_data_T,
                                self.aesara_graph.compute_formation_block(
                                    Z_x,
                                    self.aesara_graph.get_scalar_field_at_surface_points(Z_x),
                                    values_properties
                                ),
                                on_unused_input='ignore',
                                allow_input_downcast=False,
                                profile=False)

        if inplace is True:
            self.aesara_function_formation = th_fn

        if debug is True:
            print('Level of Optimization: ', aesara.config.optimizer)
            print('Device: ', aesara.config.device)
            print('Precision: ', self.dtype)
            print('Number of faults: ',
                  self.additional_data.structure_data.df.loc['values', 'number faults'])
        print('Compilation Done!')

        return th_fn

    def compile_th_fn_fault_block(self, Z_x=None, weights=None, grid=None, values_properties=None,
                                  inplace=False, debug=False):
        """

        Args:
            weights: Constant weights
            grid:  Constant grids
            inplace:
            debug:

        Returns:

        """
        self.set_aesara_shared_kriging()
        self.set_aesara_shared_structure_surfaces()

        # This are the shared parameters and the compilation of the function. This will be hidden as well at some point
        input_data_T = self.aesara_graph.input_parameters_block
        print('Compiling aesara function...')

        if weights is None:
            weights = self.aesara_graph.compute_weights()
        else:
            weights = aesara.shared(weights)

        if grid is None:
            grid = self.aesara_graph.grid_val_T
        else:
            grid = aesara.shared(grid)

        if values_properties is None:
            values_properties = self.aesara_graph.values_properties_op
        else:
            values_properties = aesara.shared(values_properties)

        if Z_x is None:
            Z_x = self.aesara_graph.compute_scalar_field(weights, grid)
        else:
            Z_x = aesara.shared(Z_x)

        th_fn = aesara.function(input_data_T,
                                self.aesara_graph.compute_fault_block(
                                    Z_x,
                                    self.aesara_graph.get_scalar_field_at_surface_points(Z_x),
                                    values_properties,
                                    0,
                                    grid
                                ),
                                # mode=NanGuardMode(nan_is_error=True),
                                on_unused_input='ignore',
                                allow_input_downcast=False,
                                profile=False)

        if inplace is True:
            self.aesara_function_faults = th_fn

        if debug is True:
            print('Level of Optimization: ', aesara.config.optimizer)
            print('Device: ', aesara.config.device)
            print('Precision: ', self.dtype)
            print('Number of faults: ',
                  self.additional_data.structure_data.df.loc['values', 'number faults'])
        print('Compilation Done!')
        return th_fn


class InterpolatorGravity:
    def set_aesara_shared_tz_kernel(self, tz=None):
        """Set the aesara component tz to each voxel"""

        if tz is None or tz == 'auto':
            try:
                tz = self.calculate_tz(self.grid.centered_grid)
            except AttributeError:
                raise AttributeError('You need to calculate or pass tz first.')

        self.aesara_graph.tz.set_value(tz.astype(self.dtype))

    def calculate_tz(self, centered_grid):
        from gempy_legacy.assets.geophysics import GravityPreprocessing
        g = GravityPreprocessing(centered_grid)

        return g.set_tz_kernel()

    def set_aesara_shared_pos_density(self, pos_density):
        self.aesara_graph.pos_density.set_value(pos_density)

    def set_aesara_shared_l0_l1(self):
        self.aesara_graph.lg0.set_value(self.grid.get_grid_args('centered')[0])
        self.aesara_graph.lg1.set_value(self.grid.get_grid_args('centered')[1])

    def set_aesara_shared_gravity(self, tz='auto', pos_density=1):
        self.set_aesara_shared_tz_kernel(tz)
        self.set_aesara_shared_pos_density(pos_density)
        self.set_aesara_shared_l0_l1()


class InterpolatorMagnetics:
    def set_aesara_shared_Vs_kernel(self, V=None):

        # Check that V is not None an is not auto consider that V is can be numpy array
        if V is not None and np.any(V != 'auto'):
           self.aesara_graph.V.set_value(V.astype(self.dtype))
        else:
            try:
                V = self.calculate_V(self.grid.centered_grid)
            except AttributeError:
                raise AttributeError('You need to calculate or pass V first.')


    def calculate_V(self, centered_grid):
        from gempy_legacy.assets.geophysics import MagneticsPreprocessing
        Vmodel = MagneticsPreprocessing(centered_grid).set_Vs_kernel()

        return Vmodel

    def set_aesara_shared_pos_magnetics(self, pos_magnetics):
        self.aesara_graph.pos_magnetics.set_value(pos_magnetics)

    def set_aesara_shared_magnetic_cts(self, incl, decl, B_ext=52819.8506939139e-9):
        """
        Args:
            B_ext : External magnetic field in [T], in magnetic surveys this is the geomagnetic field - varies temporaly
            incl  : Dip of the geomagnetic field in degrees- varies spatially
            decl  : Angle between magnetic and true North in degrees - varies spatially
        """

        self.aesara_graph.incl.set_value(incl)
        self.aesara_graph.decl.set_value(decl)
        self.aesara_graph.B_ext.set_value(B_ext)

    def set_aesara_shared_l0_l1(self):
        self.aesara_graph.lg0.set_value(self.grid.get_grid_args('centered')[0])
        self.aesara_graph.lg1.set_value(self.grid.get_grid_args('centered')[1])

    def set_aesara_shared_magnetics(self, V='auto', pos_magnetics=1,
                                    incl=None, decl=None, B_ext=52819.8506939139e-9):
        self.set_aesara_shared_Vs_kernel(V)
        self.set_aesara_shared_pos_magnetics(pos_magnetics)
        self.set_aesara_shared_magnetic_cts(incl, decl, B_ext)
        self.set_aesara_shared_l0_l1()


@_setdoc_pro(ds.ctrl)
@_setdoc([Interpolator.__doc__])
class InterpolatorModel(Interpolator, InterpolatorGravity, InterpolatorMagnetics):
    """
    Child class of :class:`Interpolator` which set the shared variables and compiles the aesara
    graph to compute the geological model, i.e. lithologies.

    Attributes:
        compute_weights_ctrl (list[bool]): [s0]
        compute_scalar_ctrl (list[bool]):
        compute_block_ctrl (list[bool]):

    Interpolator Doc
    """

    def __init__(self, surface_points: "SurfacePoints", orientations: "Orientations", grid: "Grid",
                 surfaces: "Surfaces", series, faults: "Faults", additional_data: "AdditionalData",
                 **kwargs):

        super().__init__(surface_points, orientations, grid, surfaces, series, faults,
                         additional_data, **kwargs)
        self.len_series_i = np.zeros(1)
        self.len_series_o = np.zeros(1)
        self.len_series_u = np.zeros(1)
        self.len_series_f = np.zeros(1)
        self.len_series_w = np.zeros(1)

        self.set_initial_results()

        n_series = 1000

        self.compute_weights_ctrl = np.ones(n_series, dtype=bool)
        self.compute_scalar_ctrl = np.ones(n_series, dtype=bool)
        self.compute_block_ctrl = np.ones(n_series, dtype=bool)

    def reset_flow_control_initial_results(self, reset_weights=True, reset_scalar=True,
                                           reset_block=True):
        """
        Method to reset to the initial state all the recompute ctrl. After calling this method next time
         gp.compute_model is called, everything will be computed. Panic bottom.

        Args:
            reset_weights (bool):
            reset_scalar (bool):
            reset_block (bool):

        Returns:
            True
        """
        n_series = self.len_series_i.shape[0]
        x_to_interp_shape = self.grid.values_c.shape[0] + 2 * self.len_series_i.sum()

        if reset_weights is True:
            self.compute_weights_ctrl = np.ones(1000, dtype=bool)
            self.aesara_graph.weights_vector.set_value(
                np.zeros((self.len_series_w.sum()), dtype=self.dtype))

        if reset_scalar is True:
            self.compute_scalar_ctrl = np.ones(1000, dtype=bool)
            self.aesara_graph.scalar_fields_matrix.set_value(
                np.zeros((n_series, x_to_interp_shape), dtype=self.dtype))

        if reset_block is True:
            self.compute_block_ctrl = np.ones(1000, dtype=bool)
            self.aesara_graph.mask_matrix.set_value(
                np.zeros((n_series, x_to_interp_shape), dtype='bool'))
            self.aesara_graph.block_matrix.set_value(
                np.zeros((n_series,
                          self.surfaces.df.iloc[:, self.surfaces._n_properties:].values.shape[1],
                          x_to_interp_shape), dtype=self.dtype))
        return True

    def set_flow_control(self):
        """
        Initialize the ctrl vectors to the number of series size.

        Returns:
            True
        """
        n_series = 1000
        self.compute_weights_ctrl = np.ones(n_series, dtype=bool)
        self.compute_scalar_ctrl = np.ones(n_series, dtype=bool)
        self.compute_block_ctrl = np.ones(n_series, dtype=bool)
        return True

    @_setdoc_pro(reset_flow_control_initial_results.__doc__)
    def set_all_shared_parameters(self, reset_ctrl=False):
        """
        Set all aesara shared parameters required for the computation of lithology

        Args:
            reset_ctrl (bool): If true, [s0]

        Returns:
            True
        """
        self.set_aesara_shared_loop()
        self.set_aesara_shared_relations()
        self.set_aesara_shared_kriging()
        self.set_aesara_shared_structure_surfaces()
        # self.set_aesara_shared_topology()
        if reset_ctrl is True:
            self.reset_flow_control_initial_results()

        return True

    def set_aesara_shared_topology(self):

        max_lith = self.surfaces.df.groupby('isFault')['id'].count()[False]
        if type(max_lith) != int or type(max_lith) != np.int64 or type(max_lith) != np.int32:
            max_lith = 0

        self.aesara_graph.max_lith.set_value(max_lith)
        self.aesara_graph.regular_grid_res.set_value(self.grid.regular_grid.resolution)
        self.aesara_graph.dxdydz.set_value(
            np.array(self.grid.regular_grid.get_dx_dy_dz(), dtype=self.dtype))

    @_setdoc_pro(reset_flow_control_initial_results.__doc__)
    def set_aesara_shared_structure(self, reset_ctrl=False):
        """
        Set all aesara shared variable dependent on :class:`Structure`.

        Args:
            reset_ctrl (bool): If true, [s0]

        Returns:
            True

        """
        self.set_aesara_shared_loop()
        self.set_aesara_shared_relations()
        self.set_aesara_shared_structure_surfaces()
        # universal grades
        # self.aesara_graph.n_universal_eq_T.set_value(
        #     list(self.additional_data.kriging_data.df.loc['values', 'drift equations'].astype('int32')))

        if reset_ctrl is True:
            self.reset_flow_control_initial_results()
        return True

    def remove_series_without_data(self):
        len_series_i = self.additional_data.structure_data.df.loc[
                           'values', 'len series surface_points'] - \
                       self.additional_data.structure_data.df.loc[
                           'values', 'number surfaces per series']

        len_series_o = self.additional_data.structure_data.df.loc[
            'values', 'len series orientations'].astype(
            'int32')

        # Remove series without data
        non_zero_i = len_series_i.nonzero()[0]
        non_zero_o = len_series_o.nonzero()[0]
        non_zero = np.intersect1d(non_zero_i, non_zero_o)

        self.non_zero = non_zero
        return self.non_zero

    def _compute_len_series(self):

        self.len_series_i = self.additional_data.structure_data.df.loc[
                                'values', 'len series surface_points'] - \
                            self.additional_data.structure_data.df.loc[
                                'values', 'number surfaces per series']

        self.len_series_o = self.additional_data.structure_data.df.loc[
            'values', 'len series orientations'].astype(
            'int32')

        # Remove series without data
        non_zero_i = self.len_series_i.nonzero()[0]
        non_zero_o = self.len_series_o.nonzero()[0]
        non_zero = np.intersect1d(non_zero_i, non_zero_o)

        self.non_zero = non_zero

        self.len_series_u = self.additional_data.kriging_data.df.loc[
            'values', 'drift equations'].astype('int32')
        try:
            len_series_f_ = self.faults.faults_relations_df.values[non_zero][:, non_zero].sum(
                axis=0)

        except np.AxisError:
            print('np.axis error')
            len_series_f_ = self.faults.faults_relations_df.values.sum(axis=0)

        self.len_series_f = np.atleast_1d(len_series_f_.astype(
            'int32'))  # [:self.additional_data.get_additional_data()['values']['Structure', 'number series']]

        self._old_len_series = self.len_series_i

        self.len_series_i = self.len_series_i[non_zero]
        self.len_series_o = self.len_series_o[non_zero]
        # self.len_series_f = self.len_series_f[non_zero]
        self.len_series_u = self.len_series_u[non_zero]

        if self.len_series_i.shape[0] == 0:
            self.len_series_i = np.zeros(1, dtype=int)
            self._old_len_series = self.len_series_i

        if self.len_series_o.shape[0] == 0:
            self.len_series_o = np.zeros(1, dtype=int)
        if self.len_series_u.shape[0] == 0:
            self.len_series_u = np.zeros(1, dtype=int)
        if self.len_series_f.shape[0] == 0:
            self.len_series_f = np.zeros(1, dtype=int)

        self.len_series_w = self.len_series_i + self.len_series_o * 3 + self.len_series_u + self.len_series_f

    def set_aesara_shared_loop(self):
        """Set the aesara shared variables that are looped for each series."""
        self._compute_len_series()

        self.aesara_graph.len_series_i.set_value(
            np.insert(self.len_series_i.cumsum(), 0, 0).astype('int32'))
        self.aesara_graph.len_series_o.set_value(
            np.insert(self.len_series_o.cumsum(), 0, 0).astype('int32'))
        self.aesara_graph.len_series_w.set_value(
            np.insert(self.len_series_w.cumsum(), 0, 0).astype('int32'))

        # Number of surfaces per series. The function is not pretty but the result is quite clear
        n_surfaces_per_serie = np.insert(
            self.additional_data.structure_data.df.loc['values', 'number surfaces per series'][
                self.non_zero].cumsum(), 0, 0). \
            astype('int32')
        self.aesara_graph.n_surfaces_per_series.set_value(n_surfaces_per_serie)
        self.aesara_graph.n_universal_eq_T.set_value(
            list(self.additional_data.kriging_data.df.loc['values', 'drift equations'].astype(
                'int32')[self.non_zero]))

    @_setdoc_pro(set_aesara_shared_loop.__doc__)
    def set_aesara_shared_weights(self):
        """Set the aesara shared weights and [s0]"""
        self.set_aesara_shared_loop()
        self.aesara_graph.weights_vector.set_value(
            np.zeros((self.len_series_w.sum()), dtype=self.dtype))

    def set_aesara_shared_fault_relation(self):
        self.remove_series_without_data()
        """Set the aesara shared variable with the fault relation"""
        self.aesara_graph.fault_relation.set_value(
            self.faults.faults_relations_df.values[self.non_zero][:, self.non_zero])

    def set_aesara_shared_is_fault(self):
        """Set aesara shared variable which controls if a series is fault or not"""
        is_fault_ = self.faults.df['isFault'].values[self.non_zero]
        self.aesara_graph.is_fault.set_value(is_fault_)

    def set_aesara_shared_is_finite(self):
        """Set aesara shared variable which controls if a fault is finite or not"""
        self.aesara_graph.is_finite_ctrl.set_value(self.faults.df['isFinite'].values.astype(bool))

    def set_aesara_shared_onlap_erode(self):
        """Set the aesara variables which control the masking patterns according to the uncomformity relation"""
        self.remove_series_without_data()

        is_erosion = self.series.df['BottomRelation'].values[self.non_zero] == 'Erosion'
        is_onlap = np.roll(self.series.df['BottomRelation'].values[self.non_zero] == 'Onlap', 1)

        if len(is_erosion) != 0:
            is_erosion[-1] = False
        # this comes from the series df
        self.aesara_graph.is_erosion.set_value(is_erosion)
        self.aesara_graph.is_onlap.set_value(is_onlap)

    def set_aesara_shared_faults(self):
        """Set all aesara shared variables wich controls the faults behaviour"""

        self.set_aesara_shared_fault_relation()
        # This comes from the faults df
        self.set_aesara_shared_is_fault()
        self.set_aesara_shared_is_finite()

    def set_aesara_shared_relations(self):
        """Set all aesara shared variables that control all the series interactions with each other"""
        self.set_aesara_shared_fault_relation()
        # This comes from the faults df
        self.set_aesara_shared_is_fault()
        self.set_aesara_shared_is_finite()
        self.set_aesara_shared_onlap_erode()

    def set_initial_results(self):
        """
        Initialize all the aesara shared variables where we store the final results of the interpolation.
        This function must be called always after set_aesara_shared_loop

        Returns:
            True
        """
        self._compute_len_series()

        x_to_interp_shape = self.grid.values_c.shape[0] + 2 * self.len_series_i.sum()
        n_series = self.len_series_i.shape[
            0]  # self.additional_data.structure_data.df.loc['values', 'number series']

        self.aesara_graph.weights_vector.set_value(
            np.zeros((self.len_series_w.sum()), dtype=self.dtype))
        self.aesara_graph.scalar_fields_matrix.set_value(
            np.zeros((n_series, x_to_interp_shape), dtype=self.dtype))

        self.aesara_graph.mask_matrix.set_value(
            np.zeros((n_series, x_to_interp_shape), dtype='bool'))
        self.aesara_graph.block_matrix.set_value(
            np.zeros(
                (n_series, self.surfaces.df.iloc[:, self.surfaces._n_properties:].values.shape[1],
                 x_to_interp_shape), dtype=self.dtype))
        return True

    def set_initial_results_matrices(self):
        """
        Initialize all the aesara shared variables where we store the final results of the interpolation except the
        kriging weights vector.


        Returns:
            True
        """
        self._compute_len_series()

        x_to_interp_shape = self.grid.values_c.shape[0] + 2 * self.len_series_i.sum()
        n_series = self.len_series_i.shape[
            0]  # self.additional_data.structure_data.df.loc['values', 'number series']

        self.aesara_graph.scalar_fields_matrix.set_value(
            np.zeros((n_series, x_to_interp_shape), dtype=self.dtype))

        self.aesara_graph.mask_matrix.set_value(
            np.zeros((n_series, x_to_interp_shape), dtype='bool'))
        self.aesara_graph.block_matrix.set_value(
            np.zeros(
                (n_series, self.surfaces.df.iloc[:, self.surfaces._n_properties:].values.shape[1],
                 x_to_interp_shape), dtype=self.dtype))

    def set_aesara_shared_grid(self, grid=None):
        if grid == 'shared':
            grid_sh = self.grid.values_c
            self.aesara_graph.grid_val_T = aesara.shared(grid_sh.astype(self.dtype),
                                                         'Constant values to interpolate.')
        elif grid is not None:
            self.aesara_graph.grid_val_T = aesara.shared(grid.astype(self.dtype),
                                                         'Constant values to interpolate.')

    def modify_results_matrices_pro(self):
        """
        Modify all aesara shared matrices to the right size according to the structure data. This method allows
        to change the size of the results without having the recompute all series"""

        old_len_i = self._old_len_series
        new_len_i = self.additional_data.structure_data.df.loc[
                        'values', 'len series surface_points'] - \
                    self.additional_data.structure_data.df.loc[
                        'values', 'number surfaces per series']
        if new_len_i.shape[0] < old_len_i.shape[0]:
            self.set_initial_results()
            old_len_i = old_len_i[old_len_i != 0]
        elif new_len_i.shape[0] > old_len_i.shape[0]:
            self.set_initial_results()
            new_len_i = new_len_i[new_len_i != 0]
        else:
            scalar_fields_matrix = self.aesara_graph.scalar_fields_matrix.get_value()
            mask_matrix = self.aesara_graph.mask_matrix.get_value()
            block_matrix = self.aesara_graph.block_matrix.get_value()

            len_i_diff = new_len_i - old_len_i
            for e, i in enumerate(len_i_diff):
                loc = self.grid.values_c.shape[0] + old_len_i[e]
                i *= 2
                if i == 0:
                    pass
                elif i > 0:
                    self.aesara_graph.scalar_fields_matrix.set_value(
                        np.insert(scalar_fields_matrix, [loc], np.zeros(i), axis=1))
                    self.aesara_graph.mask_matrix.set_value(np.insert(
                        mask_matrix, [loc], np.zeros(i, dtype=self.dtype), axis=1))
                    self.aesara_graph.block_matrix.set_value(np.insert(
                        block_matrix, [loc], np.zeros(i, dtype=self.dtype), axis=2))

                else:
                    self.aesara_graph.scalar_fields_matrix.set_value(
                        np.delete(scalar_fields_matrix, np.arange(loc, loc + i, -1) - 1, axis=1))
                    self.aesara_graph.mask_matrix.set_value(
                        np.delete(mask_matrix, np.arange(loc, loc + i, -1) - 1, axis=1))
                    self.aesara_graph.block_matrix.set_value(
                        np.delete(block_matrix, np.arange(loc, loc + i, -1) - 1, axis=2))

        self.modify_results_weights()

    def modify_results_weights(self):
        """Modify the aesara shared weights vector according to the structure.
        """
        old_len_w = self.len_series_w
        self._compute_len_series()
        new_len_w = self.len_series_w
        if new_len_w.shape[0] != old_len_w[0]:
            self.set_initial_results()
        else:
            weights = self.aesara_graph.weights_vector.get_value()
            len_w_diff = new_len_w - old_len_w
            for e, i in enumerate(len_w_diff):
                #   print(len_w_diff, weights)
                if i == 0:
                    pass
                elif i > 0:
                    self.aesara_graph.weights_vector.set_value(
                        np.insert(weights, old_len_w[e], np.zeros(i)))
                else:
                    #      print(np.delete(weights, np.arange(old_len_w[e],  old_len_w[e] + i, -1)-1))
                    self.aesara_graph.weights_vector.set_value(
                        np.delete(weights, np.arange(old_len_w[e], old_len_w[e] + i, -1) - 1))

    def get_python_input_block(self, append_control=True, fault_drift=None):
        """
        Get values from the data objects used during the interpolation:
             - dip positions XYZ
             - dip angles
             - azimuth
             - polarity
             - surface_points coordinates XYZ

        Args:
            append_control (bool): If true append the ctrl vectors to the input list
            fault_drift (Optional[np.array]): matrix with per computed faults to drift the model

        Returns:
            list: list of arrays with all the input parameters to the aesara function
        """
        # orientations, this ones I tile them inside aesara. PYTHON VAR
        dips_position = self.orientations.df[['X_c', 'Y_c', 'Z_c']].values
        dip_angles = self.orientations.df["dip"].values
        azimuth = self.orientations.df["azimuth"].values
        polarity = self.orientations.df["polarity"].values
        surface_points_coord = self.surface_points.df[['X_c', 'Y_c', 'Z_c']].values
        grid = self.grid.values_c
        if fault_drift is None:
            fault_drift = np.zeros((0, grid.shape[0] + 2 * self.len_series_i.sum()))

        # values_properties = np.array([[]], dtype='float32')
        # g = self.surfaces.df.groupby('series')
        # for series_ in self.series.df.index.values[self.non_zero]:
        #     values_properties = np.append(values_properties,
        #                                   g.get_group(series_).iloc[:, self.surfaces._n_properties:].values.
        #                                   astype(self.dtype).T, axis=1)

        #  values_properties = self.surfaces.df.iloc[:, self.surfaces._n_properties:].values.astype(self.dtype).T

        values_properties = self.surfaces.df.groupby('isActive').get_group(
            True).iloc[:, self.surfaces._n_properties:].values.astype(self.dtype).T
        # Set all in a list casting them in the chosen dtype
        idl = [np.cast[self.dtype](xs) for xs in (dips_position, dip_angles, azimuth, polarity,
                                                  surface_points_coord,
                                                  fault_drift, grid, values_properties)]
        if append_control is True:
            idl.append(self.compute_weights_ctrl)
            idl.append(self.compute_scalar_ctrl)
            idl.append(self.compute_block_ctrl)

        return idl

    def print_aesara_shared(self):
        """Print many of the aesara shared variables"""

        print('len sereies i', self.aesara_graph.len_series_i.get_value())
        print('len sereies o', self.aesara_graph.len_series_o.get_value())
        print('len sereies w', self.aesara_graph.len_series_w.get_value())
        print('n surfaces per series', self.aesara_graph.n_surfaces_per_series.get_value())
        print('n universal eq', self.aesara_graph.n_universal_eq_T.get_value())
        print('is finite', self.aesara_graph.is_finite_ctrl.get_value())
        print('is erosion', self.aesara_graph.is_erosion.get_value())
        print('is onlap', self.aesara_graph.is_onlap.get_value())

    def compile_th_fn_geo(self, inplace=False, debug=True, grid: Union[str, np.ndarray] = None):
        """
        Compile and create the aesara function which can be evaluated to compute the geological models

        Args:

            inplace (bool): If true add the attribute aesara.function to the object inplace
            debug (bool): If true print some of the aesara flags
            grid: If None, grid will be passed as variable. If shared or np.ndarray the grid will be treated as
             constant (if shared the grid will be taken of grid)

        Returns:
            aesara.function: function that computes the whole interpolation
        """

        self.set_all_shared_parameters(reset_ctrl=False)
        # This are the shared parameters and the compilation of the function. This will be hidden as well at some point
        input_data_T = self.aesara_graph.input_parameters_loop
        print('Compiling aesara function...')
        if grid == 'shared' or grid is not None:
            self.set_aesara_shared_grid(grid)

        th_fn = aesara.function(inputs=input_data_T,
                                outputs=self.aesara_graph.aesara_output(),
                                updates=[
                                    (self.aesara_graph.block_matrix, self.aesara_graph.new_block),
                                    (self.aesara_graph.weights_vector,
                                     self.aesara_graph.new_weights),
                                    (self.aesara_graph.scalar_fields_matrix,
                                     self.aesara_graph.new_scalar),
                                    (self.aesara_graph.mask_matrix, self.aesara_graph.new_mask)
                                ],
                                on_unused_input='ignore',
                                allow_input_downcast=False,
                                profile=False)

        if inplace is True:
            self.aesara_function = th_fn

        if debug is True:
            print('Level of Optimization: ', aesara.config.optimizer)
            print('Device: ', aesara.config.device)
            print('Precision: ', aesara.config.floatX)
            print('Number of faults: ',
                  self.additional_data.structure_data.df.loc['values', 'number faults'])
        print('Compilation Done!')

        return th_fn



from gempy_legacy.plot import visualization_2d as vv
import numpy as np
import matplotlib.patches as patches
import matplotlib.pyplot as plt

def _extract_boundaries(self, axes, section_name='topography'):
    """
    Should be part of viz 2d.
    Args:
        self:
        axes:
        section_name:

    Returns:

    """
    cs = []

    faults = list(self.model._faults.df[self.model._faults.df['isFault'] == True].index)

    if section_name == 'topography':
        shape = self.model._grid.topography.resolution
        a = self.model.solutions.geological_map[1]
        extent = [self.model._grid.topography.extent[0],
                  self.model._grid.topography.extent[1],
                  self.model._grid.topography.extent[2],
                  self.model._grid.topography.extent[3]]
    else:
        l0, l1 = self.model._grid.sections.get_section_args(section_name)
        j = np.where(self.model._grid.sections.names == section_name)[0][0]
        shape = [self.model._grid.sections.resolution[j][0], self.model._grid.sections.resolution[j][1]]
        a = self.model.solutions.sections[1][:, l0:l1]
        # b = self.model.solutions.sections[0][:, l0:l1].reshape(shape).T
        extent = [0, self.model._grid.sections.dist[j][0], self.model._grid.regular_grid.extent[4],
                  self.model._grid.regular_grid.extent[5]]

    zorder = 2
    counter = a.shape[0]

    counters = np.arange(0, counter, 1)
    c_id = 0  # color id startpoint
    colors = []
    for f_id in counters:
        block = a[f_id]
        level = self.model.solutions.scalar_field_at_surface_points[f_id][np.where(
            self.model.solutions.scalar_field_at_surface_points[f_id] != 0)]

        levels = np.insert(level, 0, block.max())
        c_id2 = c_id + len(level)
        if f_id == counters.max():
            levels = np.insert(levels, level.shape[0], block.min())
            c_id2 = c_id + len(levels)  # color id endpoint
        if section_name == 'topography':
            block = block.reshape(shape)
        else:
            block = block.reshape(shape).T
        zorder = zorder - (f_id + len(level))

        if f_id >= len(faults):
            color = self.cmap.colors[c_id:c_id2][::-1]
            plot = axes.contourf(block, 0, levels=np.sort(levels), colors=color,
                                 linestyles='solid', origin='lower',
                                 extent=extent, zorder=zorder)
        else:
            color = self.cmap.colors[c_id:c_id2][0]
            plot = axes.contour(block, 0, levels=np.sort(levels), colors=color,
                                linestyles='solid', origin='lower',
                                extent=extent, zorder=zorder)
        c_id += len(level)
        cs.append(plot)
        if type(color) == str:
            colors.append(color)
        else:
            for c in color:
                colors.append(c)
    return cs, colors, extent


def get_polygon_dictionary(geo_model, section_name):
    """

    Args:
        geo_model: the geological model
        section_name: the section from which the polygons should be retrieved. Must be 'topography' or a predefined
        section of model.grid.sections

    Returns: [0]: pathdict. A dictionary of every surface with its corresponding polygon xy values.
             [1]: color dictionary
             [2]: extent of the section

    """
    p = vv.Plot2D(geo_model)
    p.create_figure((13, 13))
    t = p.add_section(section_name, ax_pos=111)

    cs, colors, extent = _extract_boundaries(p, p.axes[0], section_name)
    all_paths = []
    for contour in cs:
        for col in contour.collections:
            all_paths.append(col.get_paths())

    for p in all_paths:
        if p == []:
            all_paths.remove(p)

    surflist = []
    for color in colors:
        surflist.append(geo_model._surfaces.df[geo_model._surfaces.df['color'] == color]['surface'].values[0])

    pathdict = dict(zip(surflist, all_paths))
    cdict = dict(zip(surflist, colors))

    return pathdict, cdict, extent




"""
The aim of this module is to encapsulate the loading functionality. Also this will enable
 better error handling when some of the data files are missing
"""
import os
import pathlib
import shutil

import pandas as pn
import numpy as np

# region Save
from gempy_legacy import Project, create_model, init_data
from gempy_legacy.utils.meta import _setdoc


@_setdoc(Project.save_model_pickle.__doc__)
def save_model_to_pickle(model: Project, path=None):
    model.save_model_pickle(path)
    return True


@_setdoc(Project.save_model.__doc__)
def save_model(model: Project, name=None, path=None, compress=True,
               solution=False, **kwargs):
    name, path = default_path_and_name(model, name, path)
    model.save_model(name, path, compress)
    if solution is True:       
        solution_to_netcdf(kwargs, model, name, path)
    if compress is True:
        shutil.make_archive(name, 'zip', path)
        shutil.rmtree(path)
    return True


def default_path_and_name(model, name, path):
    if name is None:
        name = model.meta.project_name
    if not path:
        path = './'
    path = f'{path}/{name}'
    if os.path.isdir(path):
        print("Directory already exists, files will be overwritten")
    else:
        os.makedirs(f'{path}')
    return name, path


def solution_to_netcdf(kwargs, model, name, path):
    try:
        model.solutions.to_netcdf(path, name, **kwargs)
    except AttributeError:
        raise AttributeError('You need to install Subsurface to be able to '
                             'write the solutions.')


@_setdoc(Project.load_model_pickle.__doc__)
def load_model_pickle(path):
    """
    Read InputData object from python pickle.

    Args:
       path (str): path where save the pickle

    Returns:
        :class:`Project`

    """
    return Project.load_model_pickle(path)


def load_model(name=None, path=None, recompile=False):
    """
    Loading model saved with model.save_model function.

    Args:
        name: name of folder with saved files
        path (str): path to folder directory or the zip file
        recompile (bool): if true, aesara functions will be recompiled

    Returns:
        :class:`Project`

    """
    # TODO: Divide each dataframe in its own function and move them as
    #  method of the class
    # TODO: Include try except in case some of the datafiles is missing
    #

    # Default path
    is_compressed = False
    if path is None:
        path = f'./{name}'
    p = pathlib.Path(path)

    # If the path includes .zip
    if p.suffix == '.zip':
        is_compressed, path = _unpack_model_if_compressed_includes_zip(is_compressed, path)

    # if the path does not include .zip but exist
    elif os.path.isfile(f'{path}.zip'):
        is_compressed = _unpack_model_if_compressed_no_zip(is_compressed, path)

    # create model with extent and resolution from csv - check
    geo_model = create_model()
    init_data(
        geo_model,
        np.load(f'{path}/{name}_extent.npy'),
        np.load(f'{path}/{name}_resolution.npy')
    )

    _load_files_into_geo_model(geo_model, name, path)

    if recompile is True:
        from gempy_legacy.api_modules.setters import set_interpolator
        set_interpolator(geo_model, verbose=[0])

    # Cleaning temp files
    if is_compressed:
        shutil.rmtree(path)

    return geo_model


def _load_files_into_geo_model(geo_model, name, path):
    _load_topography(geo_model, name, path)
    _load_additional_data(geo_model, name, path)
    # Load series
    cat_series = _load_stack(geo_model, name, path)
    # do surfaces properly
    cat_surfaces = _load_surfaces(cat_series, geo_model, name, path)
    # do orientations properly, reset all dtypes
    _load_orientations(cat_series, cat_surfaces, geo_model, name, path)
    # do surface_points properly, reset all dtypes
    _load_surface_points(cat_series, cat_surfaces, geo_model, name, path)
    # update structure from loaded input
    _update_structure_and_mapping(geo_model)


def _update_structure_and_mapping(geo_model):
    geo_model._additional_data.structure_data.update_structure_from_input()
    geo_model._rescaling.rescale_data()
    geo_model.update_from_series()
    geo_model.update_from_surfaces()
    geo_model.update_structure()


def _load_surface_points(cat_series, cat_surfaces, geo_model, name, path):
    geo_model._surface_points.df = pn.read_csv(f'{path}/{name}_surface_points.csv',
                                               index_col=0,
                                               dtype={'X': 'float64', 'Y': 'float64',
                                                      'Z': 'float64',
                                                      'X_r': 'float64',
                                                      'Y_r': 'float64',
                                                      'Z_r': 'float64',
                                                      'surface': 'category',
                                                      'series': 'category',
                                                      'id': 'int64',
                                                      'order_series': 'int64'})
    geo_model._surface_points.df['surface'] = geo_model._surface_points.df['surface'].cat.set_categories(cat_surfaces)
    geo_model._surface_points.df['series'] = geo_model._surface_points.df['series'].cat.set_categories(cat_series)
    # Code to add smooth columns for models saved before gempy 2.0bdev4
    try:
        geo_model._surface_points.df['smooth']
    except KeyError:
        geo_model._surface_points.df['smooth'] = 1e-7


def _load_orientations(cat_series, cat_surfaces, geo_model, name, path):
    geo_model._orientations.df = pn.read_csv(f'{path}/{name}_orientations.csv',
                                             index_col=0,
                                             dtype={'X': 'float64', 'Y': 'float64',
                                                    'Z': 'float64',
                                                    'X_r': 'float64',
                                                    'Y_r': 'float64',
                                                    'Z_r': 'float64',
                                                    'dip': 'float64',
                                                    'azimuth': 'float64',
                                                    'polarity': 'float64',
                                                    'surface': 'category',
                                                    'series': 'category',
                                                    'id': 'int64',
                                                    'order_series': 'int64'})
    geo_model._orientations.df['surface'] = geo_model._orientations.df['surface'].cat.set_categories(cat_surfaces)
    geo_model._orientations.df['series'] = geo_model._orientations.df['series'].cat.set_categories(cat_series)

    try:
        geo_model._orientations.df['smooth']
    except KeyError:
        geo_model._orientations.df['smooth'] = 0.01


def _load_surfaces(cat_series, geo_model, name, path):
    surf_df = pn.read_csv(f'{path}/{name}_surfaces.csv', index_col=0,
                          dtype={'surface': 'str', 'series': 'category',
                                 'order_surfaces': 'int64', 'isBasement': 'bool',
                                 'id': 'int64',
                                 'color': 'str'})
    c_ = surf_df.columns[
        ~(surf_df.columns.isin(geo_model._surfaces._columns_vis_drop))]
    geo_model._surfaces.df[c_] = surf_df[c_]
    geo_model._surfaces.df['series'] = geo_model._surfaces.df['series'].cat.reorder_categories(
        np.asarray(geo_model._stack.df.index),
        ordered=False)
    geo_model._surfaces.sort_surfaces()
    geo_model._surfaces.colors.generate_colordict()
    geo_model._surfaces.df['series'] = geo_model._surfaces.df['series'].cat.set_categories(cat_series)
    try:
        geo_model._surfaces.df['isActive']
    except KeyError:
        geo_model._surfaces.df['isActive'] = False
    cat_surfaces = geo_model._surfaces.df['surface'].values
    return cat_surfaces


def _load_stack(geo_model, name, path):
    s = pn.read_csv(f'{path}/{name}_series.csv', index_col=0,
                    dtype={'order_series': 'int32',
                           'BottomRelation': 'category'})
    f = pn.read_csv(f'{path}/{name}_faults.csv', index_col=0,
                    dtype={'isFault': 'bool', 'isFinite': 'bool'})
    stack = pn.concat([s, f], axis=1, sort=False)
    stack = stack.loc[:, ~stack.columns.duplicated()]
    geo_model._stack.df = stack
    series_index = pn.CategoricalIndex(geo_model._stack.df.index.values)
    # geo_model.series.df.index = pn.CategoricalIndex(series_index)
    geo_model._stack.df.index = series_index
    geo_model._stack.df['BottomRelation'] = geo_model._stack.df['BottomRelation'].cat.set_categories(
        ['Erosion', 'Onlap', 'Fault'])
    try:
        geo_model._stack.df['isActive']
    except KeyError:
        geo_model._stack.df['isActive'] = False
    cat_series = geo_model._stack.df.index.values
    # # do faults relations properly - this is where I struggle
    geo_model._faults.faults_relations_df = pn.read_csv(
        f'{path}/{name}_faults_relations.csv', index_col=0)
    geo_model._faults.faults_relations_df.index = series_index
    geo_model._faults.faults_relations_df.columns = series_index
    geo_model._faults.faults_relations_df.fillna(False, inplace=True)
    return cat_series


def _load_additional_data(geo_model, name, path):
    geo_model._additional_data.kriging_data.df = pn.read_csv(
        f'{path}/{name}_kriging_data.csv', index_col=0,
        dtype={'range': 'float64', '$C_o$': 'float64',
               'drift equations': object,
               'nugget grad': 'float64',
               'nugget scalar': 'float64'})
    geo_model._additional_data.kriging_data.str2int_u_grade()
    geo_model._additional_data.options.df = pn.read_csv(f'{path}/{name}_options.csv',
                                                        index_col=0,
                                                        dtype={'dtype': 'category',
                                                               'output': 'category',
                                                               'aesara_optimizer': 'category',
                                                               'device': 'category',
                                                               'verbosity': object})
    geo_model._additional_data.options.df['dtype'] = geo_model._additional_data.options.df['dtype'].cat.set_categories(
        ['float32', 'float64'])
    geo_model._additional_data.options.df['aesara_optimizer'] = geo_model._additional_data.options.df['aesara_optimizer'].cat.set_categories(
        ['fast_run', 'fast_compile'])
    geo_model._additional_data.options.df['device'] = geo_model._additional_data.options.df['device'].cat.set_categories(
        ['cpu', 'cuda'])
    geo_model._additional_data.options.df['output'] = geo_model._additional_data.options.df['output'].cat.set_categories(
        ['geology', 'gradients'])
    geo_model._additional_data.options.df.loc['values', 'verbosity'] = None


def _load_topography(geo_model, name, path):
    try:
        geo_model.set_topography(source='saved',
                                 filepath=f'{path}/{name}_topography.npy')
    except FileNotFoundError:
        pass


def _unpack_model_if_compressed_no_zip(is_compressed, path):
    try:
        shutil.unpack_archive(path + '.zip', extract_dir=path)
    except ValueError as e:
        raise ValueError(e)
    is_compressed = True
    return is_compressed


def _unpack_model_if_compressed_includes_zip(is_compressed, path):
    path = path[:-4]
    print("is path", path)
    try:
        shutil.unpack_archive(path + '.zip', extract_dir=path)
    except ValueError as e:
        raise ValueError(e)
    is_compressed = True
    return is_compressed, path

# endregion



"""

    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.

Module with classes and methods to perform kriging of elements (and at some point exploit the potential field to
choose the directions of the variograms)

Tested on Ubuntu 16

Created on 1/5/2017

@author: Miguel de la Varga
"""

import aesara
import aesara.tensor  as T
import matplotlib.pyplot as plt
import pymc as pm
import numpy as np
import pandas as pn

from bokeh.io import show
import bokeh.layouts as bl
import bokeh.plotting as bp


def choose_lithology_elements(df, litho, elem=None, coord = True):
    """
    litho(str): Name of the lithology-domain
    elem(list): List of strings with elements you want to analyze
    """
    # Choosing just the opx litology
    if elem is not None:
        if coord:
            domain = df[df['Lithology'] == litho][np.append(['X', 'Y', 'Z'], elem)]
        else:
            domain = df[df['Lithology'] == litho][elem]
        # Drop negative values
        domain = domain[(domain[elem] > 0).all(1)]
    else:
        domain = df[df['Lithology'] == litho][['X', 'Y', 'Z']]

    return domain


def select_segmented_grid(df, litho, grid, block):

    block = np.squeeze(block)
    assert grid.shape[0] == block.shape[0], 'The grid you want to use for kriging and the grid used for the layers ' \
                                            'segmentation are not the same'

    litho_num = df['Lithology Number'][df["Lithology"] == litho].iloc[0]
    segm_grid = grid[block == litho_num]
    return segm_grid


def transform_data(df_o, n_comp=1, log10=False):
    """
    Method to improve the normality of the input data before perform krigin
    Args:
        df_o: Dataframe with the data to interpolate
        n_comp: Number of component in case of multimodal data
        log10 (bool): If true return the log in base 10 of the properties:

    Returns:
        pandas.core.frame.DataFrame: Data frame with the transformed data
    """

    import copy
    df = copy.deepcopy(df_o)

    # Take log to try to aproximate better the normal distributions
    if log10:
        print('computing log')
        df[df.columns.difference(['X', 'Y', 'Z'])] = np.log10(df[df.columns.difference(['X', 'Y', 'Z'])])

    # Finding n modes in the data
    if n_comp > 1:
        from sklearn import mixture
        gmm = mixture.GaussianMixture(n_components=n_comp,
                                      covariance_type='full').fit(df[df.columns.difference(['X', 'Y', 'Z'])])

        # Adding the categories to the pandas frame
        labels_all = gmm.predict(df[df.columns.difference(['X', 'Y', 'Z'])])
        df['cluster'] = labels_all
    return df


def aesara_sed():
    """
    Function to create a aesara function to compute the euclidian distances efficiently
    Returns:
        aesara.compile.function_module.Function: Compiled function

    """

    aesara.config.compute_test_value = "ignore"

    # Set symbolic variable as matrix (with the XYZ coords)
    coord_T_x1 = T.dmatrix()
    coord_T_x2 = T.dmatrix()

    # Euclidian distances function
    def squared_euclidean_distances(x_1, x_2):
        sqd = T.sqrt(T.maximum(
            (x_1 ** 2).sum(1).reshape((x_1.shape[0], 1)) +
            (x_2 ** 2).sum(1).reshape((1, x_2.shape[0])) -
            2 * x_1.dot(x_2.T), 0
        ))
        return sqd

    # Compiling function
    f = aesara.function([coord_T_x1, coord_T_x2],
                        squared_euclidean_distances(coord_T_x1, coord_T_x2),
                        allow_input_downcast=False)
    return f


# This is extremily ineficient. Try to vectorize it in aesara, it is possible to gain X100
def compute_variogram(df, properties, euclidian_distances, tol=10, lags=np.logspace(0, 2.5, 100), plot=[]):
    """
    Compute the experimental variogram and cross variogram for a par of properties
    Args:
        df (pandas.core.frame.DataFrame): Dataframe with the properties and coordinates used in the experimental
         variogram computation
        properties (list): List of the two properties to compute the semivariogram.
        euclidian_distances (numpy.array): Precomputed distances of the euclidian distances
        tol (float): Tolerance
        lags (list): List of lags to compute the experimental variogram
        plot (bool): If true plot the experimental variogram after computed

    Returns:
        list: semvariance aor cross-semivariance
    """

    # Tiling the properties to a square matrix
    element = (df[properties[0]].values.reshape(-1, 1) -
               np.tile(df[properties[1]], (df[properties[1]].shape[0], 1))) ** 2

    # Semivariance computation
    semivariance_lag = []

    # Looping every lag to compute the semivariance
    for i in lags:
        # Selecting the points at the given lag and tolerance
        points_at_lag = ((euclidian_distances > i - tol) * (euclidian_distances < i + tol))

        # Extracting the values of the properties of the selected lags
        var_points = element[points_at_lag]

        # Appending the semivariance
        semivariance_lag = np.append(semivariance_lag, np.mean(var_points) / 2)

    if "experimental" in plot:
        # Visualizetion of the experimental variogram
        plt.plot(lags, semivariance_lag, 'o')

    return semivariance_lag


def exp_lags(max_range, exp=2, n_lags=100):
    """
    Function to create a more specific exponential distance between the lags in case that log10 gives too much weight
    to the smaller lags
    Args:
        max_range(float): Maximum distance
        exp (float): Exponential degree
        n_lags (int): Number of lags

    Returns:
        list: lags

    """
    lags = np.empty(0)
    for i in range(n_lags):
        lags = np.append(lags, i ** exp)
    lags = lags / lags.max() * max_range
    return lags


def compute_crossvariogram(df, properties_names, euclidian_distances=None, **kwargs):
    """
    Compute the experimental crossvariogram of all properties given
    Args:
        df (pandas.core.frame.DataFrame): Dataframe with the properties and coordinates used in the experimental
         variogram computation
        properties_names (list str): List of strings with the properties to compute
        euclidian_distances (numpy.array): Precomputed euclidian distances. If None they are computed inplace
        Keyword Args:
            - lag_exp: Degree of the exponential. If None log10
            - lag_range: Maximum distance to compute a lag
            - n_lags: Number of lags

    Returns:
        pandas.core.frame.DataFrame: Every experimental cross-variogram
    """

    lag_exp = kwargs.get('lag_exp', None)
    lag_range = kwargs.get('lag_range', 500)
    n_lags = kwargs.get('n_lags', 100)

    # Choosing the lag array
    if lag_exp is not None:
        lags = exp_lags(lag_range, lag_exp, n_lags)
    else:
        lags = np.logspace(0, np.log10(lag_range), n_lags)

    # Compute euclidian distance
    if not euclidian_distances:
        euclidian_distances = aesara_sed()(df[['X', 'Y', 'Z']], df[['X', 'Y', 'Z']])

    # Init dataframe to store the results
    experimental_variograms_frame = pn.DataFrame()

    # This is extremily ineficient. Try to vectorize it in aesara, it is possible to gain X100
    # Nested loop. DEPRECATED enumerate
    for i in properties_names:
        for j in properties_names:
            col_name = i + '-' + j
            values = compute_variogram(df, [i, j], euclidian_distances, lags=lags)
            experimental_variograms_frame[col_name] = values

    # Add lags column for plot mainly
    experimental_variograms_frame['lags'] = lags

    return experimental_variograms_frame


class SGS(object):
    """
    Class that contains the necessary methods to perform a sequential gaurssian simmulation from experimental variograms
    """
    def __init__(self, exp_var, properties=None,  n_exp=2, n_gauss=2,
                 data=None, grid=None, grid_to_inter=None, lithology=None, geomodel=None):
        """

        Args:
            exp_var (pandas.core.frame.DataFrame): Dataframe with the experimental variograms. Columns are the variables
            while raws the lags
            properties (list strings): Variables to be used
            n_exp (int): number of exponential basic functions
            n_gauss (int): number of gaussian basic functions
        """

        # Analytical covariance
        # ---------------------
        self.exp_var_raw = exp_var
        if not properties:
            self.properties = exp_var.columns[['ppm' in i for i in exp_var.columns]]
        else:
            self.properties = properties
        self.n_properties = len(self.properties)
        self.lags = self.exp_var_raw['lags']
        self.exp_var, self.nuggets = self.preprocess()

        self.n_exp = n_exp
        self.n_gauss = n_gauss
        self.trace = None

        # Kriging
        # -------
        self.data = data
        self.data_to_inter = None
        self.grid = grid
        self.grid_to_inter = grid_to_inter
        self.lithology = lithology
        self.geomodel = geomodel

        # aesara functions
        # ----------------
        self.SED_f = aesara_sed()

        self._recursion_check = 0

    def select_segmented_grid(self, grid=None, df=None, litho=None, block=None):
        """
        Extract from the full model the points belonging to a given domain (litho)
        Args:
            grid (numpy.arrau): Full model grid
            df (pandas.core.frame.DataFrame): with all data
            litho (str): Name of the lithology to interpolate
            block (numpy.array): 3D block model generated by gempy

        Returns:

        """

        # If we do not pass the attributes but are already a property of the object extract it
        if grid is None:
            grid = self.grid

        if block is not None:
            block = np.squeeze(block)
        else:
            block = self.geomodel

        if df is None:
            df = self.data

        if litho is None:
            litho = self.lithology

        # Check that the attributes have value
        assert all([any(block), ~df.empty, litho, np.any(grid)]), "Some of the attributes are not provided"
        assert grid.shape[0] == block.shape[0], 'The grid you want to use for kriging and the grid used for the layers ' \
                                                'segmentation are not the same'

        # Translate lithology name to lithology number, i.e. the value of the lithology in the gempy block
        litho_num = df['Lithology Number'][df["Lithology"] == litho].iloc[0]

        # Extract from the grid the XYZ coordinates where the gempy block model is the given lithology
        segm_grid = grid[block == litho_num]

        # Set the segmented grid
        self.grid_to_inter = segm_grid

        return segm_grid

    def choose_lithology_elements(self, df=None, litho=None, elem=None, coord=True):
        """
        Choose from the whole input data, those points which fall into the domain of interest
        litho(str): Name of the lithology-domain
        elem(list): List of strings with elements you want to analyze
        """

        # If we do not pass the attributes but are already a property of the object extract it
        if not df:
            df = self.data

        if not litho:
            litho = self.lithology

        # Check that the attributes have value
        assert all([~df.empty, litho]), "Some of the attributes are not provided"

        # Choosing just one lithology
        if elem is not None:
            if coord:
                domain = df[df['Lithology'] == litho][np.append(['X', 'Y', 'Z'], elem)]
            else:
                domain = df[df['Lithology'] == litho][elem]
            # Drop negative values
            domain = domain[(domain[elem] > 0).all(1)]
        else:
            domain = df[df['Lithology'] == litho][['X', 'Y', 'Z']]

        # Set the segmented data
        self.data_to_inter = domain

        return domain

    def set_data(self, data):
        """
        Data setter
        Args:
            data (pandas.core.frame.DataFrame): Dataframe with coordinates and values of the properties of interest
        """
        self.data = data

    def set_grid_to_inter(self, grid):
        """
        Grid setter
        Args:
            grid (numpy.array): Grid object from gempy with the points of interest of the model
        """
        self.grid_to_inter = grid

    def set_lithology(self, lithology):
        """
        Lithology setter
        Args:
            lithology (str): Name of the domain or lithology of interest
        """
        self.lithology = lithology

    def set_geomodel(self, geomodel):
        """
        gempy model setter
        Args:
            geomodel (numpy.array): gempy block model
        """
        self.geomodel = geomodel

    def set_trace(self, trace):
        """
        Analytical covarinace trace setter
        Args:
            trace (pymc3.trace): trace with the sill, range and weights of each property
        """
        self.trace = trace

    def preprocess(self):
        """
        Normalization of data between 0 and 1 and subtraction of the nuggets
        Returns:
            pandas.core.frame.DataFrame: Dataframe containing the transformed data
            pandas.core.frame.DataFrame: Containing the substracted nuggets

        """
        import sklearn.preprocessing as skp

        # Normalization
        scaled_data = pn.DataFrame(skp.minmax_scale(self.exp_var_raw[self.properties]), columns=self.properties)

        # Nuggets
        nuggets = scaled_data[self.properties].iloc[0]
        processed_data = scaled_data - nuggets
        return processed_data, nuggets

    def plot_experimental(self, transformed=False):
        """
        Plotting the experimental data either transformed (see preprocess doc) or not
        Args:
            transformed (bool): if true plot the transformed data

        Returns:
            matplotlib.plot: plot of experimental variogram
        """

        if transformed:
            plot = self.exp_var.plot(x=self.lags, y=self.properties, subplots=True, kind ='line',
                                     style='.',
                                     layout=(int(np.sqrt(self.n_properties)), int(np.sqrt(self.n_properties))),
                                     figsize=(16, 8))
        else:
            plot = self.exp_var_raw.plot(x=self.lags, y=self.properties, subplots=True, kind='line',
                                         style='.',
                                         layout=(int(np.sqrt(self.n_properties)), int(np.sqrt(self.n_properties))),
                                         figsize=(16, 8))
        return plot

    def fit_cross_cov(self, n_exp=2, n_gauss=2, range_mu=None):
        """
        Fit an analytical covariance to the experimental data.
        Args:
            n_exp (int): number of exponential basic functions
            n_gauss (int): number of gaussian basic functions
            range_mu: prior mean of the range. Default mean of the lags

        Returns:
            pymc.Model: PyMC3 model to be sampled using MCMC
        """
        self.n_exp = n_exp
        self.n_gauss = n_gauss
        n_var = self.n_properties
        df = self.exp_var
        lags = self.lags

        # Prior standard deviation for the error of the regression
        prior_std_reg = df.std(0).max() * 10

        # Prior value for the mean of the ranges
        if not range_mu:
            range_mu = lags.mean()

        # pymc3 Model
        with pm.Model() as model:  # model specifications in PyMC3 are wrapped in a with-statement
            # Define priors
            sigma = pm.HalfCauchy('sigma', beta=prior_std_reg, testval=1., shape=n_var)

            psill = pm.Normal('sill', prior_std_reg, sd=.5 * prior_std_reg, shape=(n_exp + n_gauss))
            range_ = pm.Normal('range', range_mu, sd=range_mu * .3, shape=(n_exp + n_gauss))

            lambda_ = pm.Uniform('weights', 0, 1, shape=(n_var * (n_exp + n_gauss)))

            # Exponential covariance
            exp = pm.Deterministic('exp',
                                   # (lambda_[:n_exp*n_var]*
                                   psill[:n_exp] *
                                   (1. - T.exp(T.dot(-lags.values.reshape((len(lags), 1)),
                                                     (range_[:n_exp].reshape((1, n_exp)) / 3.) ** -1))))

            gauss = pm.Deterministic('gaus',
                                     psill[n_exp:] *
                                     (1. - T.exp(T.dot(-lags.values.reshape((len(lags), 1)) ** 2,
                                                       (range_[n_exp:].reshape((1, n_gauss)) * 4 / 7.) ** -2))))

            # We stack the basic functions in the same matrix and tile it to match the number of properties we have
            func = pm.Deterministic('func', T.tile(T.horizontal_stack(exp, gauss), (n_var, 1, 1)))

            # We weight each basic function and sum them
            func_w = pm.Deterministic("func_w", T.sum(func * lambda_.reshape((n_var, 1, (n_exp + n_gauss))), axis=2))

            for e, cross in enumerate(df.columns):
                # Likelihoods
                pm.Normal(cross + "_like", mu=func_w[e], sd=sigma[e], observed=df[cross].values)
        return model

    def plot_cross_variograms(self, iter_plot=200, trace=None, experimental=None):
        """
        Plot the analytical cross-variogram of a given MCMC inference
        Args:
            iter_plot (int): Number of traces to plot
            trace (pymc3.trace): trace with the sill, range and weights of each property
            experimental (bool): if True plot the experimental variogram as well

        Returns:
            None
        """

        if not trace:
            trace = self.trace
        assert trace, 'set the trace to the object'

        n_exp = self.n_exp
        n_gauss = self.n_gauss
        lags = self.lags
        # DEP- n_equations = trace['weights'].shape[1]
        n_iter = trace['weights'].shape[0]
        lags_tiled = np.tile(lags, (iter_plot, 1))
        b_var = []
        for i in range(0, self.n_properties):  # DEP- n_equations, (n_exp+n_gaus)):
            # Init tensor
            b = np.zeros((len(lags), n_iter, 0))
            for i_exp in range(0, n_exp):

                b = np.dstack((b, trace['weights'][:, i_exp + i * (n_exp + n_gauss)] *
                               exp_vario(lags, trace['sill'][:, i_exp], trace['range'][:, i_exp])))
            for i_gauss in range(n_exp, n_gauss + n_exp):
                b = np.dstack((b, trace['weights'][:, i_gauss + i * (n_exp + n_gauss)] *
                               gaus_vario(lags, trace['sill'][:, i_gauss], trace['range'][:, i_gauss])))

            # Sum the contributins of each function
            b_all = b.sum(axis=2)
            # Append each variable
            b_var.append(b_all[:, -iter_plot:].T)

        # Bokeh code to plot this
        p_all = []
        for e, el in enumerate(self.properties):
            p = bp.figure()#x_axis_type="log")
            p.multi_line(list(lags_tiled), list(b_var[e]), color='olive', alpha=0.08)
            if experimental:
                p.scatter(self.lags, y=self.exp_var[el], color='navy', size=2)
            p.title.text = el
            p.xaxis.axis_label = "lags"
            p.yaxis.axis_label = "Semivariance"

            p_all = np.append(p_all, p)

        grid = bl.gridplot(list(p_all), ncols=5, plot_width=250, plot_height=150)

        show(grid)

    def exp_vario(self, lags, sill, range_):
        """
        Vectorial computation of exponential variogram
        Args:
            lags (numpy.array): Distances
            sill (numpy.array): Array of sills. The shape will be number of properties by number of exponential
             basic functions
            range_ (numpy.array): Array of ranges. The shape will be number of properties by number of exponential
             basic functions

        Returns:
            numpy.array: Exponential variogram for every lag and every sill and range.
        """
        return sill * (1 - np.exp(np.dot(-lags.reshape(-1, 1) * 3, range_.reshape(1, -1) ** -1)))

    def gaus_vario(self, lags, sill, range_):
        """
        Vectorial computation of gauss variogram
        Args:
            lags (numpy.array): Distances
            sill (numpy.array): Array of sills. The shape will be number of properties by number of gauss
             basic functions
            range_ (numpy.array): Array of ranges. The shape will be number of properties by number of gauss
             basic functions

        Returns:
            numpy.array: gauss variogram for every lag and every sill and range.
        """
        return sill * (1 - np.exp(np.dot(-lags.reshape(-1, 1) ** 2, (range_.reshape(1, -1) * 4 / 7) ** -2)))

    def plot_cross_covariance(self, nuggets=False, iter_plot=200):
        """
        Plot the cross covariance for the given properties
        Args:
            nuggets (numpy.array): subtracted nuggets
            iter_plot (int): number of traces to plot

        Returns:
            None
        """
        n_exp = self.n_exp
        n_gauss = self.n_gauss
        trace = self.trace
        lags = self.lags

        n_equations = trace['weights'].shape[1]
        n_iter = trace['weights'].shape[0]
        lags_tiled = np.tile(lags, (iter_plot, 1))
        b_var = []
        for i in range(0, self.n_properties):  # n_equations, (n_exp+n_gaus)):
            # Init tensor
            b = np.zeros((len(lags), n_iter, 0))
            for i_exp in range(0, n_exp):
                # print(i_exp, "exp")
                b = np.dstack((b, trace['weights'][:, i_exp + i * (n_exp + n_gauss)] *
                               exp_vario(lags, trace['sill'][:, i_exp], trace['range'][:, i_exp])))
            for i_gaus in range(n_exp, n_gauss + n_exp):
                # print(i_gaus)
                b = np.dstack((b, trace['weights'][:, i_gaus + i * (n_exp + n_gauss)] *
                               gaus_vario(lags, trace['sill'][:, i_gaus], trace['range'][:, i_gaus])))
            # Sum the contributins of each function
            if nuggets:
                b_all = 1 - (b.sum(axis=2) + self.nuggets[i])
            else:
                b_all = 1 - (b.sum(axis=2))
            # Append each variable
            b_var.append(b_all[:, -iter_plot:].T)

        p_all = []
        for e, el in enumerate(self.properties):
            p = bp.figure(x_axis_type="log")
            p.multi_line(list(lags_tiled), list(b_var[e]), color='olive', alpha=0.08)

            p.title.text = el
            p.xaxis.axis_label = "lags"
            p.yaxis.axis_label = "Semivariance"

            p_all = np.append(p_all, p)

        grid = bl.gridplot(list(p_all), ncols=5, plot_width=250, plot_height=150)

        show(grid)

    def solve_kriging(self, selection_A, selection_b, #selected_coord_data, selected_values_data, selected_grid_to_inter, trace,
                      nuggets=None, n_var=1, n_exp=2, n_gauss=2):
        """
        Solve the kriging system for n given point of the grid_to_inter property selected by selection_b using n
        points of data_to_inter given by selection_A
        Args:
            selection_A (bool): input points used in A matrix
            selection_b (bool): points to interpolate from the grid
            nuggets (numpy.array): Nugget effect of each cross variogram. If None take property
            n_var: number of variables
            n_exp (int): number of exponential basic functions
            n_gauss (int): number of gaussian basic functions

        Returns:

        """

        # Select input data and compute its euclidean distances
        selected_coord_data = self.data_to_inter[selection_A][['X', 'Y', 'Z']]
        selected_values_data = self.data_to_inter[selection_A][self.data_to_inter.columns.difference(['X', 'Y', 'Z'])].values
        h_A = self.SED_f(selected_coord_data, selected_coord_data)

        # Select points of grid to interpolate and compute the euclidean distances respect the input data
        selected_grid_to_inter = self.grid_to_inter[selection_b]
        h_b = self.SED_f(selected_coord_data, selected_grid_to_inter)

        # Parameters for the covariances
        trace = self.trace
        nuggets = self.nuggets
        n_var = int(np.sqrt(self.n_properties))
        n_exp = self.n_exp
        n_gauss = self.n_gauss

        # Choose a random trace to conserve the covariance uncertainty of the regression
        sample = np.random.randint(trace['weights'].shape[0] - 1000, trace['weights'].shape[0])

       # Compute cross-covariances
        cov_h = cross_covariance(trace, h_A, sample=sample,
                                 nuggets=nuggets, n_var=n_var, n_exp=n_exp, n_gaus=n_gauss, ordinary=True)
        cov_b = cross_covariance(trace, h_b, sample=sample, nuggets=nuggets, n_var=n_var, n_exp=n_exp,
                                 n_gaus=n_gauss,
                                 ordinary=True)

        # Solve kriging system
        k_weights = np.linalg.solve(cov_h, cov_b)

        # Number of points to interpolate
        npti = selected_grid_to_inter.shape[0]

        # Repeat the input data for every point to interpolate
        svd_tmp = np.tile(np.repeat(selected_values_data, npti, axis=1), (n_var, 1))

        # Sol ordinary kriging mean
        k_mean = (svd_tmp * k_weights[:-n_var]).sum(axis=0)

        # Sol ordinary kriging std
        k_std = svd_tmp.std(axis=0) - (k_weights * cov_b)[:-n_var].sum(axis=0) +\
                (k_weights * cov_b)[-n_var:].sum(axis=0)

        assert all(k_std) > -10, "A standard deviation of kringing is really off. Check nothing is wrong"
        # Set negatives to 0
        k_std[k_std < 0] = 0.1

        # Check the results make sense else take another sample and recompute
        import scipy
        l_low, l_high = scipy.stats.norm.interval(.95, loc=np.mean(selected_values_data, axis=0),
                                                  scale=np.std(selected_values_data, axis=0))

        if not np.all((k_mean > l_low) * (k_mean < l_high)):

            k_mean, k_std, cov_h, cov_b, k_weights, sample = self.solve_kriging(selection_A, selection_b)
            self._recursion_check += 1
            assert self._recursion_check<500, 'Too many recursions. Probably something goes wrong'

        else:
            self._recursion_check = 0
            values_interp = np.random.normal(k_mean, k_std)

        return k_mean, k_std, cov_h, cov_b, k_weights, sample  # , k_std# - svd_tmp.std(axis=0)


def fit_cross_cov(df, lags, n_exp=2, n_gaus=2, range_mu=None):
    n_var = df.columns.shape[0]
    n_basis_f = n_var * (n_exp + n_gaus)
    prior_std_reg = df.std(0).max() * 10
    #
    if not range_mu:
        range_mu = lags.mean()

    # Because is a experimental variogram I am not going to have outliers
    nugget_max = df.values.max()
    # print(n_basis_f, n_var*n_exp, nugget_max, range_mu, prior_std_reg)
    # pymc3 Model
    with pm.Model() as model:  # model specifications in PyMC3 are wrapped in a with-statement
        # Define priors
        sigma = pm.HalfCauchy('sigma', beta=prior_std_reg, testval=1., shape=n_var)

        psill = pm.Normal('sill', prior_std_reg, sd=.5 * prior_std_reg, shape=(n_exp + n_gaus))
        range_ = pm.Normal('range', range_mu, sd=range_mu * .3, shape=(n_exp + n_gaus))
        #  nugget = pm.Uniform('nugget', 0, nugget_max, shape=n_var)

        lambda_ = pm.Uniform('weights', 0, 1, shape=(n_var * (n_exp + n_gaus)))

        # Exponential covariance
        exp = pm.Deterministic('exp',
                               # (lambda_[:n_exp*n_var]*
                               psill[:n_exp] *
                               (1. - T.exp(T.dot(-lags.values.reshape((len(lags), 1)),
                                                 (range_[:n_exp].reshape((1, n_exp)) / 3.) ** -1))))

        gaus = pm.Deterministic('gaus',
                                psill[n_exp:] *
                                (1. - T.exp(T.dot(-lags.values.reshape((len(lags), 1)) ** 2,
                                                  (range_[n_exp:].reshape((1, n_gaus)) * 4 / 7.) ** -2))))

        func = pm.Deterministic('func', T.tile(T.horizontal_stack(exp, gaus), (n_var, 1, 1)))

        func_w = pm.Deterministic("func_w", T.sum(func * lambda_.reshape((n_var, 1, (n_exp + n_gaus))), axis=2))
        #           nugget.reshape((n_var,1)))

        for e, cross in enumerate(df.columns):
            # Likelihoods
            pm.Normal(cross + "_like", mu=func_w[e], sd=sigma[e], observed=df[cross].values)
    return model


def exp_vario(lags, sill, range_):
    return sill * (1 - np.exp(np.dot(-lags.reshape(-1, 1) * 3, range_.reshape(1, -1) ** -1)))


def gaus_vario(lags, sill, range_):
    return sill * (1 - np.exp(np.dot(-lags.reshape(-1, 1) ** 2, (range_.reshape(1, -1) * 4 / 7) ** -2)))


def plot_cross_variograms(trace, lags, df, n_exp=2, n_gaus=2, iter_plot=200, experimental=None):
    n_equations = trace['weights'].shape[1]
    n_iter = trace['weights'].shape[0]
    lags_tiled = np.tile(lags, (iter_plot, 1))
    b_var = []
    for i in range(0, df.shape[1]):  # n_equations, (n_exp+n_gaus)):
        # Init tensor
        b = np.zeros((len(lags), n_iter, 0))
        for i_exp in range(0, n_exp):
            # print(i_exp, "exp")
            b = np.dstack((b, trace['weights'][:, i_exp + i * (n_exp + n_gaus)] *
                           exp_vario(lags, trace['sill'][:, i_exp], trace['range'][:, i_exp])))
        for i_gaus in range(n_exp, n_gaus + n_exp):
            # print(i_gaus)
            b = np.dstack((b, trace['weights'][:, i_gaus + i * (n_exp + n_gaus)] *
                           gaus_vario(lags, trace['sill'][:, i_gaus], trace['range'][:, i_gaus])))
        # Sum the contributins of each function
        b_all = b.sum(axis=2)
        # Append each variable
        b_var.append(b_all[:, -iter_plot:].T)

    p_all = []
    for e, el in enumerate(df.columns):
        p = bp.figure(x_axis_type="log")
        p.multi_line(list(lags_tiled), list(b_var[e]), color='olive', alpha=0.08)
        if experimental is not None:
            p.scatter(experimental['lags'], y=experimental[el], color='navy', size=2)
        p.title.text = el
        p.xaxis.axis_label = "lags"
        p.yaxis.axis_label = "Semivariance"

        p_all = np.append(p_all, p)

    grid = bl.gridplot(list(p_all), ncols=5, plot_width=200, plot_height=150)

    show(grid)


def plot_cross_covariance(trace, lags, df, n_exp=2, n_gaus=2, nuggets=None, iter_plot=200):
    n_equations = trace['weights'].shape[1]
    n_iter = trace['weights'].shape[0]
    lags_tiled = np.tile(lags, (iter_plot, 1))
    b_var = []
    for i in range(0, df.shape[1]):  # n_equations, (n_exp+n_gaus)):
        # Init tensor
        b = np.zeros((len(lags), n_iter, 0))
        for i_exp in range(0, n_exp):
            # print(i_exp, "exp")
            b = np.dstack((b, trace['weights'][:, i_exp + i * (n_exp + n_gaus)] *
                           exp_vario(lags, trace['sill'][:, i_exp], trace['range'][:, i_exp])))
        for i_gaus in range(n_exp, n_gaus + n_exp):
            # print(i_gaus)
            b = np.dstack((b, trace['weights'][:, i_gaus + i * (n_exp + n_gaus)] *
                           gaus_vario(lags, trace['sill'][:, i_gaus], trace['range'][:, i_gaus])))
        # Sum the contributins of each function
        if nuggets is not None:
            b_all = 1 - (b.sum(axis=2) + nuggets[i])
        else:
            b_all = 1 - (b.sum(axis=2))
        # Append each variable
        b_var.append(b_all[:, -iter_plot:].T)

    p_all = []
    for e, el in enumerate(df.columns):
        p = bp.figure(x_axis_type="log")
        p.multi_line(list(lags_tiled), list(b_var[e]), color='olive', alpha=0.08)

        p.title.text = el
        p.xaxis.axis_label = "lags"
        p.yaxis.axis_label = "Semivariance"

        p_all = np.append(p_all, p)

    grid = bl.gridplot(list(p_all), ncols=5, plot_width=250, plot_height=150)

    show(grid)


def cross_covariance(trace, sed, sample, nuggets=None, n_var=1, n_exp=2, n_gaus=2, ordinary=True):
    """

    Args:
        trace:
        sed:
        nuggets:
        n_var:
        n_exp:
        n_gaus:
        ordinary:

    Returns:

    """
    h = np.ravel(sed)
    n_points = len(h)
    n_points_r = sed.shape[0]
    n_points_c = sed.shape[1]
    #sample = np.random.randint(trace['weights'].shape[0]-200, trace['weights'].shape[0])
    n_eq = trace['weights'].shape[1]
    # Exp contribution
    exp_cont = (np.tile(
        exp_vario(h, trace['sill'][sample][:n_exp], trace['range'][sample][:n_exp]),
        n_var**2
    ) * trace['weights'][sample][np.linspace(0, n_eq-1, n_eq) % (n_exp + n_gaus) < n_exp]).reshape(n_points, n_exp, n_var**2, order="F")

    # Gauss contribution
    gaus_cont = (np.tile(
        gaus_vario(h, trace['sill'][sample][n_exp:], trace['range'][sample][n_exp:]),
        n_var**2
    ) * trace['weights'][sample][np.linspace(0, n_eq-1, n_eq) % (n_exp + n_gaus) >= n_exp]).reshape(n_points, n_gaus, n_var**2, order="F")

    # Stacking and summing
    conts = np.hstack((exp_cont, gaus_cont)).sum(axis=1)

    if nuggets is not None:
        conts += nuggets

    cov = 1 - conts

    cov_str = np.zeros((0, n_points_c*n_var))

    cov_aux = cov.reshape(n_points_r, n_points_c, n_var**2, order='F')

    # This is a incredibly convoluted way to reshape the cross covariance but I did not find anything better
    for i in range(n_var):
        cov_str = np.vstack((cov_str,
                             cov_aux[:, :, i*n_var:(i+1)*n_var].reshape(n_points_r, n_points_c*n_var, order='F')))

    if ordinary:
        ord = np.zeros((n_var, n_var*n_points_c+n_var))
        for i in range(n_var):
            ord[i, n_points_c * i:n_points_c * (i + 1)] = np.ones(n_points_c)
        cov_str = np.vstack((cov_str, ord[:, :-n_var]))

        # Stack the ordinary values to the C(h)
        if n_points_r == n_points_c:
             cov_str = np.hstack((cov_str, ord.T))

    return cov_str


def clustering_grid(grid_to_inter, n_clusters=50, plot=False):
    from sklearn.cluster import KMeans
    clust = KMeans(n_clusters=n_clusters).fit(grid_to_inter)
    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        ax.scatter(clust.cluster_centers_[:, 0], clust.cluster_centers_[:, 1], clust.cluster_centers_[:, 2])

        ax.set_xlabel('X Label')
        ax.set_ylabel('Y Label')
        ax.set_zlabel('Z Label')

        plt.show()

    return clust


def select_points(df, grid_to_inter, cluster, SED_f = aesara_sed(), n_rep=10):

    points_cluster = np.bincount(cluster.labels_)
  #  SED_f = aesara_sed()

    for i in range(n_rep):
        for i_clust in range(cluster.n_clusters):
            cluster_bool = cluster.labels_ == i_clust
            cluster_grid = grid_to_inter[cluster_bool]
            # Mix the values of each cluster
            if i is 0:
                np.random.shuffle(cluster_grid)

            size_range = int(points_cluster[i_clust]/n_rep)
            selected_cluster_grid = cluster_grid[i * size_range:(i + 1) * size_range]
            dist = SED_f(df, selected_cluster_grid)

            # Checking the radius of the simulation
            for r in range(100, 1000, 100):
                select = (dist < r).any(axis=1)
                if select.sum() > 50:
                    break

            h_x0 = dist
            yield (h_x0, select, selected_cluster_grid)


def SGS_compute_points(selected_coord_data, selected_grid_to_inter, selected_values_data,
                trace, nuggets=None, n_var=1, n_exp=2, n_gaus=2):
    #SED_f = aesara_sed()


    #SED = SED_f(selected_coord_data, selected_coord_data)

    npti = selected_grid_to_inter.shape[1]

    cov_h = cross_covariance(trace, selected_coord_data,
                             nuggets=nuggets, n_var=n_var, n_exp=n_exp, n_gaus=n_gaus, ordinary=True)
    cov_b = cross_covariance(trace, selected_grid_to_inter, nuggets=nuggets, n_var=n_var, n_exp=n_exp, n_gaus=n_gaus,
                             ordinary=True)

    k_weights = np.linalg.solve(cov_h, cov_b)
    svd_tmp = np.tile(np.repeat(selected_values_data, npti, axis=1), (n_var, 1))

    # Sol ordinary kriging mean
    k_mean = (svd_tmp * k_weights[:-n_var]).sum(axis=0)

    # Sol ordinary kriging std
    k_std = svd_tmp.std(axis=0) - (k_weights * cov_b)[:-n_var].sum(axis=0) + (k_weights * cov_b)[-n_var:].sum(axis=0)

    assert all(k_std) > -10, "A standard deviation of kringing is really off. Check nothing is wrong"

    # Set negatives to 0
    k_std[k_std < 0] = 0.1

    values_interp = np.random.normal(k_mean, k_std)
  #  for point in range(npti-1):
   #     values_data = np.vstack((values_data, values_interp[point::npti]))

    #coord_data = np.vstack((coord_data, grid_interpolating))
    return values_interp#, k_std# - svd_tmp.std(axis=0)


def SGS_run(df, grid_to_inter, cluster,
            trace, nuggets=None, n_var=1, n_exp=2, n_gaus=2,
            n_rep=10, verbose = 0):

    points_cluster = np.bincount(cluster.labels_)
    coord_data = df[['X', 'Y', 'Z']].values
    values_data = df[df.columns.difference(['X', 'Y', 'Z'])].values

    SED_f = aesara_sed()

    for i in range(n_rep):
        for i_clust in range(cluster.n_clusters):
            cluster_bool = cluster.labels_ == i_clust
            cluster_grid = grid_to_inter[cluster_bool]
            # Mix the values of each cluster
           # if i is 0:
           #     np.random.shuffle(cluster_grid)

            size_range = int(points_cluster[i_clust]/n_rep)

            # Select points where interpolate
            selected_cluster_grid = cluster_grid[i * size_range:(i + 1) * size_range]
            npti = selected_cluster_grid.shape[0]

            # Euclidiand distances
            h_x0 = SED_f(coord_data, selected_cluster_grid)
            # Drop any point already simulated
            #print((h_x0==0).sum())
          #  h_x0 = h_x0[~np.any(h_x0 == 0, axis=1)]
            h_xi = SED_f(coord_data, coord_data)

            # Checking the radius of the simulation
            for r in range(50, 1000, 1):
                select = (h_x0 < r).any(axis=1)
                if select.sum() > 50:
                    break
            if verbose > 2:
                print("sel", select.shape)
                print("val", values_data.shape)
            if verbose > 0:
                print("Number of points used and number of points to interpolate", select.sum(), npti)
            h_xi_sel = h_xi[select][:, select]
            h_x0_sel = h_x0[select]
            values_data_sel = values_data[select]

            values_interpolated = SGS_compute(h_xi_sel, h_x0_sel, values_data_sel,
                                              trace, nuggets, n_var, n_exp, n_gaus)

            # Append the coordinates of the interpolated values to the values coord data since they will be interpolated
            coord_data = np.vstack((coord_data, selected_cluster_grid))

            # Setting negative values to 0
            values_data[values_data < 0] = 0

            # Append interpolated values to the initial values
            for point in range(npti-1):
                values_data = np.vstack((values_data, values_interpolated[point::npti]))

    return coord_data, values_data

# def SGS(coord_data, values_data, h_x0, select,
#         trace, nuggets=None, n_var=1, n_exp=2, n_gaus=2):
#     SED_f = aesara_sed()
#
#     #for h_x0, select, grid_interpolating in selector:
#
#     SED = SED_f(coord_data, coord_data)
#
#     values_interp, npti = SGS_compute(SED[select], h_x0[select], values_data[select],
#                                       trace, nuggets, n_var, n_exp, n_gaus)
#
#     for point in range(npti-1):
#         values_data = np.vstack((values_data, values_interp[point::npti]))
#
#     coord_data = np.vstack((coord_data, grid_interpolating))




"""
GemPy - Subsurface Link
=======================
"""

import pooch
import numpy as np
import pandas as pd

import subsurface as sb
from subsurface.reader import read_netcdf

data_url = "https://raw.githubusercontent.com/softwareunderground/subsurface/main" \
           "/examples/tutorials/wells_unstructured.nc"

data_hash = "206290db4e563e379361725349ebf4a02628f4700d361599aedff37fab9cf5b9"
borehole_unstructured_data_file = pooch.retrieve(url=data_url,
                                                 known_hash=data_hash)

unstruct = read_netcdf.read_unstruct(borehole_unstructured_data_file)
unstruct

# %%
element = sb.LineSet(unstruct)
lines_mesh = sb.visualization.to_pyvista_line(element, radius=50)

# Plot default LITH
sb.visualization.pv_plot([lines_mesh])


# %% md
# Findig the boreholes bases
# --------------------------
# GemPy interpolates the bottom of a unit, therefore we need to be able to extract those points to be able tointerpolate them.
# xarray, pandas and numpy are using the same type of memory representation what makes possible to use the same or at least similar methods to manipulate the data to our will.
# Lets find the base points of each well:

# %%

# Creating references to the xarray.DataArray
cells_attr = unstruct.data.cell_attrs
cells = unstruct.data.cells
vertex = unstruct.data.vertex
# Find vertex points at the boundary of two units
# Marking each vertex
bool_prop_change = cells_attr.values[1:] != cells_attr.values[:-1]
# Getting the index of the vertex
args_prop_change = np.where(bool_prop_change)[0]
# Getting the attr values at those points
vals_prop_change = cells_attr[args_prop_change]
vals_prop_change.to_pandas()

# %%
# Getting the vertex values at those points
vertex_args_prop_change = cells[args_prop_change, 1]
interface_points = vertex[vertex_args_prop_change]
interface_points


# %%
# Creating a new UnstructuredData
interf_us= sb.UnstructuredData.from_array(vertex=interface_points.values, cells="points",
                                          cells_attr=vals_prop_change.to_pandas())
interf_us

# %% md
# This new UnstructuredData object instead containing data that represent lines, contain point data at the bottom of each
# unit. We can plot it very similar as before:

element = sb.PointSet(interf_us)
point_mesh = sb.visualization.to_pyvista_points(element)
sb.visualization.pv_plot([lines_mesh, point_mesh])

# %% md
# GemPy: Initialize model
# -----------------------
# The first step to create a GemPy model is create a gempy.

# %%

import gempy_legacy as gp
geo_model = gp.create_model("getting started")
geo_model.set_regular_grid(extent=[275619, 323824, 3914125, 3961793, -3972.6, 313.922], resolution=[50,50,50])
gp.set_interpolator(geo_model, aesara_optimizer='fast_compile', verbose=[])

# %% md
# Making a model step by step.
# ----------------------------

# The temptation at this point is to bring all the points into gempy and just interpolate. However, often that strategy
# results in ill posed problems due to noise or irregularities in the data. gempy has been design to being able to
# iterate rapidly and therefore a much better workflow use to be creating the model step by step.
#
# To do that, lets define a function that we can pass the name of the formation and get the assotiated vertex. Grab from
# the interf_us the XYZ coordinates of the first layer:

# %%
def get_interface_coord_from_surfaces(surface_names: list, verbose=False):
    df = pd.DataFrame(columns=["X", "Y", "Z", "surface"])

    for e, surface_name in enumerate(surface_names):
        # The properties in subsurface start at 1
        val_property = formations.index(surface_name) + 1
        # Find the cells with the surface id
        args_from_first_surface = np.where(vals_prop_change == val_property)[0]
        if verbose: print(args_from_first_surface)
        # Find the vertex
        points_from_first_surface = interface_points[args_from_first_surface]
        if verbose: print(points_from_first_surface)

        # xarray.DataArray to pandas.DataFrame
        surface_pandas = points_from_first_surface.to_pandas()

        # Add formation column
        surface_pandas["surface"] = surface_name
        df = df.append(surface_pandas)

    return df.reset_index()

# %% md
# Surfaces
# ++++++++

formations = ["topo", "etchegoin", "macoma", "chanac", "mclure",
              "santa_margarita", "fruitvale",
              "round_mountain", "olcese", "freeman_jewett", "vedder", "eocene",
              "cretaceous",
              "basement", "null"]

# %%
geo_model.add_features("Formations")
one_formation_every = 3
geo_model.add_surfaces(formations[0:4*one_formation_every:one_formation_every])

geo_model.map_stack_to_surfaces({"Formations": ["etchegoin", "macoma", "chanac", "mclure"],
                                 "Default series": ["topo"]},
                                set_series=False)


# %%
gempy_surface_points = get_interface_coord_from_surfaces(formations[0:3*one_formation_every:one_formation_every])

# %%
geo_model.set_surface_points(gempy_surface_points, update_surfaces=False)
geo_model.update_to_interpolator()


# %% md
# Adding orientations
# -------------------

# %%
# find neighbours
neighbours = gp.select_nearest_surfaces_points(geo_model, geo_model._surface_points.df, 2)

# calculate all fault orientations
gp.set_orientation_from_neighbours_all(geo_model, neighbours)

# %% md
# Using the flag to subsurface, the result of the interpolation will get stored in `subsurface` data objects. In the
# future exporting to subsurface will be the default behaviour.

# %%
gp.compute_model(geo_model, to_subsurface=True)

# %%
p3d = gp.plot_3d(geo_model)

# %%
geo_model.solutions.s_regular_grid

# %%
geo_model.solutions.meshes







import pyvista as pv
import pyvistaqt as pvqt
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from gempy_legacy.plot._vista import Vista


class MainWindow(QMainWindow):  # QtWidgets.QWidget
    def __init__(self, geo_model, parent=None):
        super(MainWindow, self).__init__(parent)

        self.create_menu_bar()
        self.bar = None

        self.main_widget = MainWidget(geo_model, parent=self)
        self.setCentralWidget(self.main_widget)

    def create_menu_bar(self):
        self.bar = self.menuBar()

        file = self.bar.addMenu("File")
        _quit = QAction("Quit", self)
        file.addAction(_quit)


class MainWidget(QWidget):
    def __init__(self, geo_model, parent=None):
        super(MainWidget, self).__init__(parent)
        self.model = geo_model

        # init base ui
        hbox = QHBoxLayout(self)
        splitter = QSplitter(Qt.Horizontal)

        self.init_tree()
        splitter.addWidget(self.tree)

        plot = QFrame()
        self.Vista = Vista(geo_model,
                           plotter_type="basic")  # init Vista plotter
        self.vtk_widget = pvqt.QtInteractor(plot)
        self.Vista.p = self.vtk_widget  # set Plotter to the vtk widget Plotter

        splitter.addWidget(self.vtk_widget)
        hbox.addWidget(splitter)
        self.setLayout(hbox)

        # self.Vista.plot_surface_points_all()

    def init_tree(self):
        self.tree = QTreeWidget()
        self.tree.setColumnCount(1)

        self.tree_items = {"surfaces": {}}
        self.tree_actors = {"surfaces": {}}

        for id_, row in self.model._surfaces.df.iterrows():
            item = QTreeWidgetItem([row.surface])
            item.setCheckState(0, Qt.Unchecked)
            self.tree.addTopLevelItem(item)
            self.tree_items["surfaces"][row.surface] = item

        self.tree.itemClicked.connect(self._check_tree_status)

    def _check_tree_status(self):
        for name, item in self.tree_items["surfaces"].items():

            if item.checkState(0) == Qt.Checked:
                actor = self.Vista.plot_surface(name)
                # self.Vista.p.renderer.add_actor(actor)
                self.tree_actors["surfaces"][name] = actor


            elif item.checkState(0) == Qt.Unchecked:
                actor = self.tree_actors["surfaces"].get(name)
                if actor:
                    self.Vista.p.remove_actor(actor)
                    # self.Vista.p.renderer.remove_actor(actor)
                    self.tree_actors["surfaces"][name] = None



"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


@author: Alexander Schaaf
"""
import numpy as np
from typing import List, Set, Tuple, Dict, Union, Optional
import matplotlib.pyplot as plt


def _get_nunconf(geo_model) -> int:
    return np.count_nonzero(
        geo_model._stack.df.BottomRelation == "Erosion"
    ) - 2  # TODO -2 n other lith series


def _get_nfaults(geo_model) -> int:
    return np.count_nonzero(geo_model._faults.df.isFault)


def _get_fb(geo_model) -> np.ndarray:
    n_unconf = _get_nunconf(geo_model)
    n_faults = _get_nfaults(geo_model)
    return np.round(
        geo_model.solutions.block_matrix[n_unconf:n_faults + n_unconf, 0, :]
    ).astype(int).sum(axis=0).reshape(*geo_model._grid.regular_grid.resolution)


def _get_lb(geo_model) -> np.ndarray:
    return np.round(
        geo_model.solutions.lith_block
    ).astype(int).reshape(*geo_model._grid.regular_grid.resolution)


def compute_topology(
        geo_model,
        cell_number: int = None,
        direction: str = None,
        n_shift: int = 1,
        voxel_threshold: int = 1
):
    """Compute topology of a GemPy model

    Args:
        geo_model (Project): GemPy model project
        cell_number (int, optional): Cell number in chosen direction. Defaults to None.
        direction (str, optional): one of the model's dimensions, x, y, or z. Defaults to None.
        n_shift (int, optional): number of voxels the model is shifted and then substracted for finding interfaces. Defaults to 1.
        voxel_threshold (int, optional): amount of voxels which have to be connected to be considered into the topology calculation. Defaults to 1.

    Returns:
        edges, centroids [numpy array]: edges and centroids of the topology graph
    """
    res = geo_model._grid.regular_grid.resolution
    fb = _get_fb(geo_model)
    lb = _get_lb(geo_model)
    n_lith = len(np.unique(lb))  # ? quicker looking it up in geomodel?

    if cell_number is None or direction is None:
        direction = "None"
    elif direction.capitalize() == "X":
        lb = lb[cell_number, :, :]
        fb = fb[cell_number, :, :]
        res = (1, res[1], res[2])
    elif direction.capitalize() == "Y":
        lb = lb[:, cell_number, :]
        fb = fb[:, cell_number, :]
        res = (res[0], 1, res[2])
    elif direction.capitalize() == "Z":
        lb = lb[:, :, cell_number]
        fb = fb[:, :, cell_number]
        res = (res[0], res[1], 1)

    edges, centroids = _analyze_topology(
        fb.ravel(),
        lb.ravel(),
        n_lith,
        res,
        n_shift,
        voxel_threshold,
        direction
    )

    edges = set((n1, n2) for n1, n2 in edges)
    edges = _filter_reverse_edges(edges)  # ? still necessary? would next line
    # ? not be enough?
    edges = _sort_edge_tuple_nodes(edges)
    return edges, centroids


def _filter_reverse_edges(edges: Set[Tuple[int, int]]) -> Set[Tuple[int, int]]:
    """Filter reversed topology edge tuples to fix doubling of topology edges
    like (1,5) (5,1).
    
    Source:
        https://stackoverflow.com/a/9922322/8040299

    Args:
        edges (Set[Tuple[int, int]]): Set of topologyedge tuples.
    
    Returns:
        Set[Tuple[int, int]]: Filtered set of topology edge tuples
    """

    edges_unique = set()
    for e in edges:
        if not (e in edges_unique or (e[1], e[0]) in edges_unique):
            edges_unique.add(e)
    return edges_unique


def _sort_edge_tuple_nodes(edges: Set[Tuple[int, int]]) -> Set[Tuple[int, int]]:
    """Sort nodes within edge tuples by ascending order.
    
    Args:
        edges (Set[Tuple[int, int]]): Set of edge tuples.
    
    Returns:
        Set[Tuple[int, int]]: Set of sorted edge tuples.
    """
    sorted_edges = set()
    for n1, n2 in edges:
        if n1 > n2:
            sorted_edges.add((int(n2), int(n1)))
        else:
            sorted_edges.add((int(n1), int(n2)))
    return sorted_edges


def _analyze_topology(
        fault_matrix_sum,
        lith_matrix,
        n_lith,
        res,
        n_shift,
        voxel_threshold,
        direction
):
    fault_shift = fault_matrix_sum.min()
    fault_matrix_sum_shift = fault_matrix_sum - fault_shift

    where = np.tile(lith_matrix, (n_lith, 1)) == np.unique(lith_matrix).reshape(
        -1, 1)
    lith_matrix_shift = np.sum(where * np.arange(n_lith).reshape(-1, 1),
                               axis=0) + 1

    topo_matrix = lith_matrix_shift + n_lith * fault_matrix_sum_shift
    topo_matrix_3D = topo_matrix.reshape(*res)

    # x direction
    if direction.capitalize() != "X":
        x_l = topo_matrix_3D[n_shift:, :, :]
        x_r = topo_matrix_3D[:-n_shift, :, :]
        x_edges, x_count = _get_edges(x_l, x_r)
        x_edges = x_edges[:, x_count > voxel_threshold]
    else:
        x_edges = np.array([[], []])

    # y direction
    if direction.capitalize() != "Y":
        y_l = topo_matrix_3D[:, n_shift:, :]
        y_r = topo_matrix_3D[:, :-n_shift, :]
        y_edges, y_count = _get_edges(y_l, y_r)
        y_edges = y_edges[:, y_count > voxel_threshold]
    else:
        y_edges = np.array([[], []])

    # z direction
    if direction.capitalize() != "Z":
        z_l = topo_matrix_3D[:, :, n_shift:]
        z_r = topo_matrix_3D[:, :, :-n_shift]
        z_edges, z_count = _get_edges(z_l, z_r)
        z_edges = z_edges[:, z_count > voxel_threshold]
    else:
        z_edges = np.array([[], []])

    edges = np.unique(
        np.concatenate((x_edges.T, y_edges.T, z_edges.T), axis=0), axis=0
    )

    centroids = _get_centroids(topo_matrix_3D)

    return edges, centroids


def get_lot_node_to_lith_id(
        geo_model,
        centroids: Dict[int, np.ndarray]
) -> Dict[int, int]:
    """Get look-up table to translate topology node id's back into GemPy lith
    id's.
    
    Args:
        geo_model: GemPy geomodel instance with solutions.
        centroids (Dict[int, Array[float, 3]]): Topology node centroids.
    
    Returns:
        Dict[int, int]: Look-up table translating node id -> lith id.
    """
    lb = geo_model.solutions.lith_block.reshape(
        geo_model._grid.regular_grid.resolution
    ).astype(int)

    lot = {}
    for node, pos in centroids.items():
        p = np.round(pos).astype(int)
        lith_id = lb[p[0], p[1], p[2]]
        lot[node] = lith_id
    return lot


def get_lot_lith_to_node_id(
        lot: Dict[int, np.ndarray]
) -> Dict[int, List[int]]:
    """Get look-up table to translate lith id's back into topology node
    id's.
    
    Args:
        lot (Dict[int, Array[float, 3]]): Node to lith id look-up table. Can be
        computed using the function 'get_lot_node_to_lith_id'.
    
    Returns:
        Dict[int, List[int]]: Look-up table.
    """
    lot2 = {}
    for k, v in lot.items():
        if v not in lot2.keys():
            lot2[v] = [k]
        else:
            lot2[v].append(k)
    return lot2


def get_lot_node_to_fault_block(
        geo_model,
        centroids: Dict[int, np.ndarray]
) -> Dict[int, int]:
    """Get a look-up table to access fault block id's for each topology node
    id.
    
    Args:
        geo_model: Geomodel instance.
        centroids (Dict[int, Array[float, 3]]): Geomodel topology centroids.
    
    Returns:
        Dict[int, int]: Look-up table.
    """
    n_lith = len(get_lith_ids(geo_model))
    lot = {}
    for node, _ in centroids.items():
        lot[node] = (node - 0 - (node // n_lith)) // n_lith
    return lot


def get_fault_ids(geo_model) -> List[int]:
    """Get fault id's of all faults in given geomodel.
    
    Args:
        geo_model: Geomodel instance
    
    Returns:
        List[int]: List of fault id's.
    """
    f_series_names = geo_model._faults.df[geo_model._faults.df.isFault].index
    fault_ids = [0]
    for fsn in f_series_names:
        fid = geo_model._surfaces.df[
                geo_model._surfaces.df.series == fsn].id.values[0]
        fault_ids.append(fid)
    return fault_ids


def get_lith_ids(geo_model, basement: bool = True) -> List[int]:
    """ Get lithology id's of all lithologies (except basement) in given
     geomodel.
    
    Args:
        geo_model: Geomodel instance.
    
    Returns:
        List[int]: List of lithology id's.
    """
    fmt_series_names = geo_model._faults.df[~geo_model._faults.df.isFault].index
    lith_ids = []
    for fsn in fmt_series_names:
        if not basement:
            if fsn == "Basement":
                continue
        lids = geo_model._surfaces.df[
            geo_model._surfaces.df.series == fsn].id.values
        for lid in lids:
            lith_ids.append(lid)
    return lith_ids


def get_detailed_labels(
        geo_model,
        edges: Set[Tuple[int, int]],
        centroids: Dict[int, np.ndarray]
) -> Tuple[Set[Tuple[str, str]], Dict[str, np.ndarray]]:
    """Convert given edges and centroids data into more detailed labels with
     pattern 'lithid_faultid'.
    
    Args:
        geo_model: [description]
        edges (Set[Tuple[int, int]]): Set of geomodel topology edges.
        centroids (Dict[int, Array[float, 3]]): Geomodel topology centroids.
    
    Returns:
        Tuple[Set[Tuple[str, str]], Dict[str, Array[float, 3]]]: Re-labeled
            edges and centroids.
    """
    lot_lith = get_lot_node_to_lith_id(geo_model, centroids)
    lot_fault = get_lot_node_to_fault_block(geo_model, centroids)

    centroids_ = {}
    for node, pos in centroids.items():
        n = lot_lith.get(node), lot_fault.get(node)
        n = str(n[0]) + "_" + str(n[1])
        centroids_[n] = pos

    edges_ = set()
    for n1, n2 in edges:
        edges_.add(
            (
                str(lot_lith.get(n1)) + "_" + str(lot_fault.get(n1)),
                str(lot_lith.get(n2)) + "_" + str(lot_fault.get(n2))
            )
        )
    return edges_, centroids_


def _get_edges(
        l: np.ndarray,
        r: np.ndarray
) -> Optional[np.ndarray]:
    """Get edges from given shifted arrays.

    Args:
        l (Array): Topology labels array shifted to one direction.
        r (Array): Topology labels array shifted to the other direction.

    Returns:
        Array: Topology edges.
    """
    shift = np.stack([l.ravel(), r.ravel()])
    i1, i2 = np.nonzero(np.diff(shift, axis=0))
    if len(i2) == 0:  # in case not edges are found (symmetric model along axis)
        return np.array([[], []]), np.array([])
    else:
        return np.unique(shift[:, i2], axis=1, return_counts=True)


def clean_unconformity_topology(
        geo_model,
        unconf_lith_id: int,
        edges: np.ndarray,
        centroids: Dict[int, np.ndarray]
) -> Tuple[Set[Tuple[int, int]], Dict[int, np.ndarray]]:
    """Clean unconformity topology edges and centroids. Needs to be run for
    each unconformity separately.
    
    Args:
        geo_model ([type]): [description]
        unconf_lith_id (int): [description]
        edges (Array[int, ..., 2]): [description]
        centroids (Dict[int, Array[int, ..., 3]]): [description]
    
    Returns:
        Tuple[Set[Tuple[int, int]], Dict[int, Array[int, ..., 3]]]: [description]
    """
    lot = get_lot_node_to_lith_id(geo_model, centroids)
    lot2 = get_lot_lith_to_node_id(lot)
    edges_clean = []

    unconf_centroid = np.mean(
        [centroids[i] for i in lot2[unconf_lith_id]], axis=0
    )
    for node_id in lot2[unconf_lith_id]:
        centroids.pop(node_id)
    centroids[lot2[unconf_lith_id][0]] = unconf_centroid

    for n1, n2 in edges:
        if n1 in lot2[unconf_lith_id]:
            n1 = lot2[unconf_lith_id][0]
        if n2 in lot2[unconf_lith_id]:
            n2 = lot2[unconf_lith_id][0]
        if n1 == n2:
            continue
        edges_clean.append((n1, n2))

    return set(edges_clean), centroids


def jaccard_index(
        edges1: Set[Tuple[int, int]],
        edges2: Set[Tuple[int, int]]
) -> float:
    """Jaccard index.
    
    Args:
        edges1 (Set[Tuple[int, int]]): Set of topology edges.
        edges2 (Set[Tuple[int, int]]): Set of topology edges.
    
    Returns:
        float: Jaccard index.
    """
    intersection_cardinality = len(edges1.intersection(edges2))
    union_cardinality = len(edges1.union(edges2))
    return intersection_cardinality / union_cardinality


def _get_centroids(labels: np.ndarray) -> dict:
    """Get geobody node centroids in array coordinates.
    
    Args:
        labels (Array[int, ..., ..., ...]): Uniquely labeled block.
    
    Returns:
        dict: Geobody node keys yield centroid coordinate tuples in array
            coordinates.
    """
    node_locs = []
    ulabels = np.unique(labels)
    for node in ulabels:
        node_pos = np.argwhere(labels == node)
        node_locs.append(node_pos.mean(axis=0))
    centroids = {n: loc for n, loc in zip(ulabels, node_locs)}
    # for k, v in centroids.items():
    # debug(f"{k}: {v}")
    return centroids


def get_adjacency_matrix(
        geo_model,
        edges: Set[Tuple[int, int]],
        centroids,
) -> np.ndarray:
    """[summary]
    
    Args:
        geo_model ([type]): [description]
        edges (Set[Tuple): [description]
    
    Returns:
        Array[bool, ..., ...]: [description]
    """
    f_ids = get_fault_ids(geo_model)
    lith_ids = get_lith_ids(geo_model)
    n = len([(l, f) for f in f_ids for l in lith_ids])

    M = np.zeros((n, n))
    lith_lot = get_lot_node_to_lith_id(geo_model, centroids)
    fault_lot = get_lot_node_to_fault_block(geo_model, centroids)
    for e1, e2 in edges:
        #     print("nodes:", e1, e2)
        l1, l2 = lith_lot.get(e1), lith_lot.get(e2)
        #     print("lith: ", l1, l2)
        f1, f2 = fault_lot.get(e1), fault_lot.get(e2)
        #     print("fault:", f1, f2)
        lp1 = np.argwhere(lith_ids == l1)[0, 0]
        lp2 = np.argwhere(lith_ids == l2)[0, 0]
        #     print("lpos :", lp1, lp2)
        p1 = lp1 + len(lith_ids) * f1
        p2 = lp2 + len(lith_ids) * f2
        #     print("pos  :", p1, p2)
        M[p1, p2] = 1
        M[p2, p1] = 1

    M = np.flip(np.flip(M, axis=1), axis=0)
    return M.astype(bool)


def _get_adj_matrix_labels(geo_model):
    f_ids = get_fault_ids(geo_model)
    lith_ids = get_lith_ids(geo_model)
    adj_matrix_labels = [(l, f) for f in f_ids for l in lith_ids]
    adj_matrix_lith_labels = [l for f in f_ids for l in lith_ids]
    adj_matrix_fault_labels = [f for f in f_ids for l in lith_ids]
    return adj_matrix_labels, adj_matrix_lith_labels, adj_matrix_fault_labels


def plot_adjacency_matrix(
        geo_model,
        adj_matrix: np.ndarray
):
    """

    Args:
        geo_model:
        adj_matrix: Array[bool, ..., ...]

    Returns:

    """
    f_ids = get_fault_ids(geo_model)
    n_faults = len(f_ids) // 2
    lith_ids = get_lith_ids(geo_model)
    n_liths = len(lith_ids)
    adj_matrix_labels, adj_matrix_lith_labels, adj_matrix_fault_labels = _get_adj_matrix_labels(
        geo_model)
    # ///////////////////////////////////////////////////////
    n = len(adj_matrix_labels)
    fig, ax = plt.subplots(figsize=(n // 2.5, n // 2.5))

    ax.imshow(adj_matrix, cmap="Greys", alpha=1)
    ax.set_xlim(-.5, n_liths * n_faults * 2 - 0.5)
    ax.set_ylim(-.5, n_liths * n_faults * 2 - 0.5)

    ax.set_title("Topology Adjacency Matrix")

    # ///////////////////////////////////////////////////////
    # lith tick labels
    ax.set_xticks(np.arange(n))
    ax.set_yticks(np.arange(n))
    ax.set_xticklabels(adj_matrix_lith_labels[::1], rotation=0)
    ax.set_yticklabels(adj_matrix_lith_labels[::1], rotation=0)

    # ///////////////////////////////////////////////////////
    # lith tick labels colors
    colors = list(geo_model._surfaces.colors.colordict.values())
    bboxkwargs = dict(
        edgecolor='none',
    )
    for xticklabel, yticklabel, l in zip(ax.xaxis.get_ticklabels(),
                                         ax.yaxis.get_ticklabels(),
                                         adj_matrix_labels[::1]):
        color = colors[l[0] - 1]

        xticklabel.set_bbox(
            dict(facecolor=color, **bboxkwargs)
        )
        xticklabel.set_color("white")

        yticklabel.set_bbox(
            dict(facecolor=color, **bboxkwargs)
        )
        yticklabel.set_color("white")

    # ///////////////////////////////////////////////////////
    # fault block tick labeling
    newax = fig.add_axes(ax.get_position())
    newax.patch.set_visible(False)

    newax.spines['bottom'].set_position(('outward', 29))
    newax.set_xlim(0, n_faults * 2)

    newax.set_xticks(np.arange(1, n_faults * 2 + 1) - 0.5)
    newax.set_xticklabels(["FB " + str(i + 1) for i in range(n_faults * 2)])

    newax.spines['left'].set_position(('outward', 25))
    newax.set_ylim(0, n_faults * 2)
    newax.set_yticks(np.arange(1, n_faults * 2 + 1) - 0.5)
    newax.set_yticklabels(
        ["FB " + str(i + 1) for i in range(n_faults * 2)][::1])

    # ///////////////////////////////////////////////////////
    # (dotted) lines for fb's
    dlinekwargs = dict(
        color="black",
        linestyle="dashed",
        alpha=0.75,
        linewidth=1
    )
    linekwargs = dict(
        color="black",
        linewidth=1
    )
    for i in range(0, n_faults * 2 + 1):
        pos = i * n_liths - .5

        if i != 0 and i != n_faults * 2:
            ax.axvline(pos, **dlinekwargs)
            ax.axhline(pos, **dlinekwargs)

        # solid spines outside to separate fbs
        line = ax.plot((-3.3, -.51), (pos, pos), **linekwargs)
        line[0].set_clip_on(False)

        line = ax.plot((pos, pos), (-3, -.51), **linekwargs)
        line[0].set_clip_on(False)
    # ///////////////////////////////////////////////////////
    return


def check_adjacency(
        edges: set,
        n1: Union[int, str],
        n2: Union[int, str]
) -> bool:
    """Check if given nodes n1 and n2 are adjacent in given topology
    edge set.
    
    Args:
        edges (set): Topology edges.
        n1 (Union[int, str]): Node 1 label.
        n2 (Union[int, str]): Node 2 label
    
    Returns:
        bool: True if adjacent, otherwise False.
    """
    if (n1, n2) in edges or (n2, n1) in edges:
        return True
    else:
        return False


def get_adjacencies(
        edges: set,
        node: Union[int, str]
) -> set:
    """Get node labels of all adjacent geobodies of geobody with given node
     in given set of edges.
    
    Args:
        edges (set): Topology edges.
        node (Union[int, str]): Node label.
    
    Returns:
        set: All adjacent geobody node labels.
    """
    adjacencies = set()
    for n1, n2 in edges:
        if node == n1:
            adjacencies.add(n2)
        elif node == n2:
            adjacencies.add(n1)
    return adjacencies


def count_unique_topologies(edges: List[Set[Tuple[int, int]]]):
    """Count unique topologie graphs in given list of edge sets.

    Args:
        edges: List of topology edge sets.
            E.g. [{(0,1), (0,2), ...}, {(0,1), (0,2), ...}]

    Returns:
        unique edges
        unique edges count
        unique edges idx
    """
    unique_edges = [edges[0]]
    unique_edges_count = [1]
    unqiue_edges_idx = [0]
    for _, topology in enumerate(edges[1:]):
        skip = False
        for b, utopology in enumerate(unique_edges):
            if utopology == topology:
                unique_edges_count[b] += 1
                unqiue_edges_idx.append(b)
                skip = True
                break
        if skip:
            continue
        unique_edges.append(topology)
        unique_edges_count.append(1)
        unqiue_edges_idx.append(len(unique_edges))

    return unique_edges, np.array(unique_edges_count), np.array(
        unqiue_edges_idx)
